/*
 * DotConverter.cpp
 *
 *  Created on: Apr 1, 2013
 *      Author: Martin Zak
 */

#include "DotConverter.h"
#include "automaton/common/State.h"

#include <automaton/FSM/NFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/PDA/NPDA.h>
#include <automaton/PDA/SinglePopNPDA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/SinglePopDPDA.h>
#include <automaton/TM/OneTapeDTM.h>

#include <exception/AlibException.h>

#include <alphabet/SymbolToStringComposer.h>
#include <regexp/RegExpToStringComposer.h>
#include <string/StringToStringComposer.h>
#include <label/LabelToStringComposer.h>

#include <set>
#include <map>
#include <list>
#include <utility>
#include <vector>
#include <typeinfo>

void DotConverter::convert(const automaton::Automaton& a, std::ostream& out) {
	a.getData().Accept((void*) &out, DotConverter::instance);
}

void DotConverter::convert(const automaton::UnknownAutomaton&, std::ostream&) {
	throw exception::AlibException("Unknown automaton is not convertible to dot.");
}

void DotConverter::convert(const automaton::EpsilonNFA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const automaton::State state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::NFA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const automaton::State state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::DFA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::ExtendedNFA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const automaton::State state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::CompactNFA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const automaton::State state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::DPDA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << a.getInitialState() << ";\n";

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::SinglePopDPDA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << a.getInitialState() << ";\n";

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::NPDA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const automaton::State state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::SinglePopNPDA& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const automaton::State state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

void DotConverter::convert(const automaton::OneTapeDTM& a, std::ostream& out) {
	label::LabelToStringComposer composer;

	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<automaton::State, int> states;
	for (const automaton::State& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const automaton::State& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << composer.compose(state.getName()) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << composer.compose(state.first.getName()) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

void DotConverter::transitions(const automaton::EpsilonNFA& fsm, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		std::string symbol;
		if (transition.first.second.is<string::Epsilon>()) {
			symbol = "&epsilon;";
		} else {
			alphabet::SymbolToStringComposer composer;
			symbol = composer.compose(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		for(const automaton::State& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ", " + symbol;
			}
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::NFA& fsm, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		alphabet::SymbolToStringComposer composer;
		std::string symbol = composer.compose(transition.first.second);

		for(const automaton::State& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ", " + symbol;
			}
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::DFA& fsm, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		alphabet::SymbolToStringComposer composer;
		std::string symbol = composer.compose(transition.first.second);

		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(transition.second)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ", " + symbol;
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::ExtendedNFA& fsm, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		regexp::RegExpToStringComposer composer;
		std::string symbol = composer.compose(transition.first.second);

		for(const automaton::State& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ", " + symbol;
			}
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::CompactNFA& fsm, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		string::StringToStringComposer composer;
		std::string symbol = composer.compose(string::String { transition.first.second } );

		for(const automaton::State& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ", " + symbol;
			}
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::DPDA& pda, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon>()) {
			symbol = "&epsilon;";
		} else {
			alphabet::SymbolToStringComposer composer;
			symbol = composer.compose(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		if (std::get<2>(transition.first).size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : std::get<2>(transition.first)) {
				alphabet::SymbolToStringComposer composer;
				symbol += " " + composer.compose(symb);
			}

		}

		symbol += " ->";

		//Push part
		if (transition.second.second.size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : transition.second.second) {
				alphabet::SymbolToStringComposer composer;
				symbol += " " + composer.compose(symb);
			}

		}

		//Insert into map
		std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(transition.second.first)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ", " + symbol;
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::SinglePopDPDA& pda, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon>()) {
			symbol = "&epsilon;";
		} else {
			alphabet::SymbolToStringComposer composer;
			symbol = composer.compose(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		alphabet::SymbolToStringComposer composer;
		symbol += " " + composer.compose(std::get<2>(transition.first));

		symbol += " ->";

		//Push part
		if (transition.second.second.size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : transition.second.second) {
				alphabet::SymbolToStringComposer composer;
				symbol += " " + composer.compose(symb);
			}

		}

		//Insert into map
		std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(transition.second.first)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ", " + symbol;
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::NPDA& pda, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon>()) {
			symbol = "&epsilon;";
		} else {
			alphabet::SymbolToStringComposer composer;
			symbol = composer.compose(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		if (std::get<2>(transition.first).size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : std::get<2>(transition.first)) {
				alphabet::SymbolToStringComposer composer;
				symbol += " " + composer.compose(symb);
			}

		}

		symbol += " ->";

		for(const auto& to : transition.second) {
			//Push part
			if (to.second.size() == 0) {
				symbol += " &epsilon;";
			} else {
				for (alphabet::Symbol symb : to.second) {
					alphabet::SymbolToStringComposer composer;
					symbol += " " + composer.compose(symb);
				}

			}

			//Insert into map
			std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(to.first)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ", " + symbol;
			}
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::SinglePopNPDA& pda, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon>()) {
			symbol = "&epsilon;";
		} else {
			alphabet::SymbolToStringComposer composer;
			symbol = composer.compose(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		alphabet::SymbolToStringComposer composer;
		symbol += " " + composer.compose(std::get<2>(transition.first));

		symbol += " ->";

		for(const auto& to : transition.second) {
			//Push part
			if (to.second.size() == 0) {
				symbol += " &epsilon;";
			} else {
				for (alphabet::Symbol symb : to.second) {
					alphabet::SymbolToStringComposer composer;
					symbol += " " + composer.compose(symb);
				}

			}

			//Insert into map
			std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(to.first)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ", " + symbol;
			}
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::OneTapeDTM& tm, const std::map<automaton::State, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;
	for (const auto& transition : tm.getTransitions()) {
		alphabet::SymbolToStringComposer composer;

		std::string symbol;

		//input symbol
		symbol = "(";
		symbol += composer.compose(transition.first.second);
		symbol += ", ";
		symbol += composer.compose(std::get<1>(transition.second));
		symbol += " ";
		switch(std::get<2>(transition.second)) {
		case automaton::Shift::LEFT:
			symbol += "&larr;";
			break;
		case automaton::Shift::RIGHT:
			symbol += "&rarr;";
			break;
		case automaton::Shift::NONE:
			symbol += "&times;";
			break;
		default:
			throw exception::AlibException("Unexpected shift direction");
		}

		//Insert into map
		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(std::get<0>(transition.second))->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ", " + symbol;
		}
	}

	//print the map
	for (const std::pair<std::pair<int, int>, std::string> transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

const DotConverter DotConverter::instance;

void DotConverter::Visit(void* data, const automaton::UnknownAutomaton& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::EpsilonNFA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::NFA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::DFA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::ExtendedNFA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::CompactNFA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::DPDA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::SinglePopDPDA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::NPDA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::SinglePopNPDA& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

void DotConverter::Visit(void* data, const automaton::OneTapeDTM& automaton) const {
	DotConverter::convert(automaton, *((std::ostream*) data));
}

