#include <iostream>
#include <fstream>
#include "automaton/FSM/FiniteAutomatonFromStringParser.h"
#include "automaton/FSM/FiniteAutomatonToStringComposer.h"
#include "grammar/GrammarFromStringParser.h"
#include "grammar/GrammarToStringComposer.h"
#include "regexp/RegExpFromStringParser.h"
#include "regexp/RegExpToStringComposer.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#include "automaton/Automaton.h"
#include "automaton/UnknownAutomaton.h"
#include "factory/DataFactory.hpp"
#include "exception/AlibException.h"

#include "DotConverter.h"
#include "GasTexConverter.h"

#define AUTOMATON_FROM_STRING 1
#define AUTOMATON_TO_STRING 2
#define AUTOMATON_TO_DOT 3
#define AUTOMATON_TO_GAS_TEX 4
#define GRAMMAR_FROM_STRING 5
#define GRAMMAR_TO_STRING 6
#define REG_EXP_FROM_STRING 7
#define REG_EXP_TO_STRING 8

void printHelp() {
	std::cout << "Help" << std::endl;
}

int automatonFromString(std::istream& in, std::ostream& out) {
	automaton::FiniteAutomatonFromStringParser parser(in);
	
	automaton::Automaton result = parser.parseValue();
	
	alib::DataFactory::toStream(result, out);
	
	return 0;
}

int grammarFromString(std::istream& in, std::ostream& out) {
	grammar::GrammarFromStringParser parser(in);
	
	grammar::Grammar result = parser.parseValue();
	
	alib::DataFactory::toStream(result, out);
	
	return 0;
}

int regexpFromString(std::istream& in, std::ostream& out) {
	regexp::RegExpFromStringParser parser(in);

	regexp::RegExp result = parser.parseValue();

	alib::DataFactory::toStream(result, out);

	return 0;
}

int automatonToString(std::istream& in, std::ostream& out) {
	try {
		automaton::Automaton automaton = alib::DataFactory::fromStream<automaton::Automaton>(in);
		automaton::FiniteAutomatonToStringComposer composer;
		out << composer.compose(automaton);
	} catch (exception::AlibException& e) {
		std::cerr << e.getCause() << std::endl;
		return 1;
	}
	return 0;
}

int grammarToString(std::istream& in, std::ostream& out) {
	try {
		grammar::Grammar grammar = alib::DataFactory::fromStream<grammar::Grammar>(in);
		grammar::GrammarToStringComposer composer;
		out << composer.compose(grammar);
	} catch (exception::AlibException& e) {
		std::cerr << e.getCause() << std::endl;
		return 1;
	}
	return 0;
}

int regexpToString(std::istream& in, std::ostream& out) {
	try {
		regexp::RegExp regexp = alib::DataFactory::fromStream<regexp::RegExp>(in);
		regexp::RegExpToStringComposer composer;
		out << composer.compose(regexp);
	} catch (exception::AlibException& e) {
		std::cerr << e.getCause() << std::endl;
		return 1;
	}
	return 0;
}

int automatonToDot(std::istream& in, std::ostream& out) {
	try {

		automaton::Automaton automaton = alib::DataFactory::fromStream<automaton::Automaton>(in);
		DotConverter::convert(automaton, out);

	} catch (exception::AlibException& e) {
		std::cerr << e.getCause() << std::endl;
		return 1;
	}
	return 0;
}

int automatonToGasTex(std::istream& in, std::ostream& out) {
	try {

		automaton::Automaton automaton = alib::DataFactory::fromStream<automaton::Automaton>(in);
		GasTexConverter::convert(automaton, out);

	} catch (exception::AlibException& e) {
		std::cerr << e.getCause() << std::endl;
		return 1;
	}
	return 0;
}

int main(int argc, char* argv[]) {

	int fromTo = 0;
	int c;
	
	while (1) {
		static struct option long_options[] = {
			{"automaton_from_string",	no_argument,	&fromTo, AUTOMATON_FROM_STRING},
			{"automaton_to_string",	 	no_argument,	&fromTo, AUTOMATON_TO_STRING},
			{"automaton_to_dot",	 	no_argument,	&fromTo, AUTOMATON_TO_DOT},
			{"automaton_to_gastex",	 	no_argument,	&fromTo, AUTOMATON_TO_GAS_TEX},
			{"grammar_from_string",		no_argument,	&fromTo, GRAMMAR_FROM_STRING},
			{"grammar_to_string",	 	no_argument,	&fromTo, GRAMMAR_TO_STRING},
			{"regexp_from_string",		no_argument,	&fromTo, REG_EXP_FROM_STRING},
			{"regexp_to_string",	 	no_argument,	&fromTo, REG_EXP_TO_STRING},
			{0, 0, 0, 0}
		};
	
		int option_index = 0;
		
		c = getopt_long (argc, argv, "v:", long_options, &option_index);

		if (c == -1)
			break;
	
		switch (c) {
		case 'v':
			printHelp();
			break;
		case 0:
			break;
		default:
			optind--;
			goto more;
		}
	}
more:

	std::istream* in = &std::cin;
	if(optind != argc) in = new std::ifstream(argv[optind]);

	int res;
	if(fromTo == AUTOMATON_FROM_STRING) {
		res = automatonFromString(*in, std::cout);
	} else if(fromTo == AUTOMATON_TO_STRING) {
		res = automatonToString(*in, std::cout);
	} else if(fromTo == AUTOMATON_TO_DOT) {
		res = automatonToDot(*in, std::cout);
	} else if(fromTo == AUTOMATON_TO_GAS_TEX) {
		res = automatonToGasTex(*in, std::cout);
	} else if(fromTo == GRAMMAR_FROM_STRING) {
		res = grammarFromString(*in, std::cout);
	} else if(fromTo == GRAMMAR_TO_STRING) {
		res = grammarToString(*in, std::cout);
	} else if(fromTo == REG_EXP_FROM_STRING) {
		res = regexpFromString(*in, std::cout);
	} else if(fromTo == REG_EXP_TO_STRING) {
		res = regexpToString(*in, std::cout);
	} else {
		printHelp();
		res = 1;
	}
	
	if(c != -1) delete in;
	return res;
}

