#include <iostream>
#include <getopt.h>
#include <string.h>
#include <ctype.h>

#include "automaton/UnknownAutomaton.h"
#include "automaton/Automaton.h"
#include "automaton/FSM/FSM.h"
#include "automaton/PDA/PDA.h"
#include "AutomatonFactory.h"
#include "AlibException.h"

#include "fsm/FsmDeterminizer.h"
#include "idpda/IdpdaDeterminizer.h"
#include "vpa/VpaDeterminizer.h"
#include "vpa/VpaDeterminizer2.h"
#include "vpa/VpaDeterminizer3.h"
#include "rhdpda/RhdpdaDeterminizer.h"
#include "rhdpda/RhdpdaDeterminizer2.h"
#include "rhdpda/RhdpdaDeterminizer3.h"
#include "rhdpda/RhdpdaDeterminizer4.h"

using namespace std;
using namespace automaton;
using namespace determinization;
using namespace alib;

#define VERSION "0.0.1"

#define TYPE_FSM "fsm"
#define TYPE_IDPDA "idpda"
#define TYPE_VPA "vpa"
#define TYPE_RHDPDA "rhdpda"

#define VERSION_1 "1"
#define VERSION_2 "2"
#define VERSION_3 "3"
#define VERSION_4 "4"


/**
 * Prints help to standard output.
 */
void printHelp()
{
    cout << "adeterminize " << VERSION << endl;
    cout << "Determinize various types of automaton." << endl;
    cout << "Usage: adeterminize -t TYPE [SWITCH...]" << endl;
    cout << "Possible arguments:" << endl;
    cout << "  -t, --type=TYPE \t Specifies type of input automaton, possible values are 'FSM' for final-state machine, 'IDPDA' for input-driven pushdown automaton, 'VPA' for visible pushdown automaton and 'RHDPDA' for real-time height deterministic pushdown automaton." << endl;
    cout << "  -a, --algorithm=VERSION_OF_ALGORITHM \t Specifies version of algorithm. This argument works only with VPA or RHDPDA type. VPA determinization has 3 versions (use numbers 1, 2 and 3) and RHDPDA determinization has 4 versions (use numbers 1, 2, 3 and 4). Default value is always number 1." << endl;
    cout << "  -h, --help \t Displays this help message." << endl;
    cout << endl;
}


/**
 * Throws exception when input automaton isn't fsm.
 *
 * @param unknownAutomaton input automaton
 * @throws AlibException when input automaton isn't fsm
 */
void validateInputAutomatonIsFsm(UnknownAutomaton& unknownAutomaton)
{
    if (!AutomatonFactory::isFSM(unknownAutomaton)) {
        throw AlibException("Input automaton must be final-state machine.");
    }
}


/**
 * Throws exception when input automaton isn't pda.
 *
 * @param unknownAutomaton input automaton
 * @throws AlibException when input automaton isn't pda
 */
void validateInputAutomatonIsPda(UnknownAutomaton& unknownAutomaton)
{
    if (!AutomatonFactory::isPDA(unknownAutomaton)) {
        throw AlibException("Input automaton must be pushdown automaton.");
    }
}


/**
 * Validates type of input automaton.
 *
 * @param unknownAutomaton input automaton
 * @throws AlibException when type of automaton isn't known or when automaton isn't of the right type
 */
void validateAutomatonType(UnknownAutomaton& unknownAutomaton, string type)
{
    if (type == TYPE_FSM) {
        validateInputAutomatonIsFsm(unknownAutomaton);
    } else if (type == TYPE_IDPDA || type == TYPE_VPA || type == TYPE_RHDPDA) {
        validateInputAutomatonIsPda(unknownAutomaton);
    } else {
        throw AlibException("Unknown type of input automaton. See help.");
    }
}


/**
 * Returns vpa determinizer according to given algorithm version.
 *
 * @param automaton input automaton
 * @param version version of algorithm
 * @return determinizer
 */
Determinizer* getVpaDeterminizer(Automaton* automaton, string version)
{
    if (version == VERSION_3) {
        return new vpa::VpaDeterminizer3((PDA*) automaton);
    } else if (version == VERSION_2) {
        return new vpa::VpaDeterminizer2((PDA*) automaton);
    }
    return new vpa::VpaDeterminizer((PDA*) automaton);
}


/**
 * Returns rhdpda determinizer according to given algorithm version.
 *
 * @param automaton input automaton
 * @param version version of algorithm
 * @return determinizer
 */
Determinizer* getRhdpdaDeterminizer(Automaton* automaton, string version)
{
    if (version == VERSION_4) {
        return new rhdpda::RhdpdaDeterminizer4((PDA*) automaton);
    } else if (version == VERSION_3) {
        return new rhdpda::RhdpdaDeterminizer3((PDA*) automaton);
    } else if (version == VERSION_2) {
        return new rhdpda::RhdpdaDeterminizer2((PDA*) automaton);
    }
    return new rhdpda::RhdpdaDeterminizer((PDA*) automaton);
}


/**
 * Converts given type to lower case.
 *
 * @param type
 * @return type in lower case
 */
string getLowerCaseType(string type)
{
    string lowerCaseType;
    for (const auto& c : type) {
        lowerCaseType.append(1, tolower(c));
    }
    return lowerCaseType;
}


/**
 * Returns determinizer according to given automaton type and algorithm version.
 *
 * @param automaton input automaton
 * @param type type of automaton
 * @param version algorithm version
 * @return determinizer
 */
Determinizer* getDeterminizer(Automaton* automaton, string type, string version)
{
    if (type == TYPE_FSM) {
        return new fsm::FsmDeterminizer((FSM*) automaton);

    } else if (type == TYPE_IDPDA) {
        return new idpda::IdpdaDeterminizer((PDA*) automaton);

    } else if (type == TYPE_VPA) {
        return getVpaDeterminizer(automaton, version);

    } else if (type == TYPE_RHDPDA) {
        return getRhdpdaDeterminizer(automaton, version);
    }

    return NULL; // should not be reached
}


int main(int argc, char** argv)
{
    static struct option longOptions[] = {
      {"type", required_argument, NULL, 't'},
      {"algorithm", required_argument, NULL, 'a'},
      {"version", no_argument, NULL, 'v'},
      {"help", no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };

    int longIndex = 0;
    int opt = 0;
    string type;
    string version = "1";

    while ((opt = getopt_long(argc, argv, "t:a:vh", longOptions, &longIndex)) != -1) {
        switch(opt) {
            case 't':
                type.assign(optarg, strlen(optarg));
                break;
            case 'a':
                version.assign(optarg, strlen(optarg));
                break;
            case 'v':
            case 'h':
            default:
                printHelp();
                return 0;
        }
    }
    
    type = getLowerCaseType(type);

    try {
        string input(istreambuf_iterator<char>(cin), (istreambuf_iterator<char>()));
        UnknownAutomaton unknownAutomaton = AutomatonFactory::fromString(input);
        validateAutomatonType(unknownAutomaton, type);
        Automaton* automaton = AutomatonFactory::buildAutomaton(unknownAutomaton);
        Determinizer* determinizer = getDeterminizer(automaton, type, version);
        Automaton* deterministicAutomaton = determinizer->determinize();
        deterministicAutomaton->toXML(cout);
        
        delete automaton;
        delete deterministicAutomaton;
        delete determinizer;
        
    } catch (AlibException& e) {
        cout << e.what() << endl;
    }

    return 0;
}
