#ifndef RHDPDADETERMINZER3_H_
#define RHDPDADETERMINZER3_H_

#include <string>
#include <set>
#include <map>
#include <vector>

#include "automaton/PDA/PDA.h"
#include "automaton/Transition.h"
#include "automaton/exception/AutomatonException.h"
#include "alphabet/Symbol.h"

#include "../common/Utils.h"
#include "../Determinizer.h"
#include "RhdpdaUtils.h"
#include "RhdpdaStructs.h"

using namespace std;
using namespace automaton;
using namespace alphabet;
using namespace determinization;

namespace determinization {
namespace rhdpda {

/**
 * Class for running third version of determinization algorithm on rhdpda.
 */
class RhdpdaDeterminizer3 : public Determinizer
{

    private:

        /** Real-time height-deterministic pushdown automaton */
        PDA* rhdpda;

        /** Real-time deterministic pushdown automaton */
        PDA* rdpda;

        /** Map of states of deterministic pda, where key is the name of state and value is data about this state */
        map<string, StateData> states;

        /** Set of internal transitions from rhdpda */
        set<TransitionPDA> internalTransitions;

        /** Set of push transitions from rhdpda */
        set<TransitionPDA> pushTransitions;

        /** Set of pop transitions from rhdpda */
        set<TransitionPDA> popTransitions;

        /** Set of all possible pairs of states and stack symbols */
        set<StateSymbolPair> allStateSymbolPairs;

        /** Vector with all possible S components of deterministic pda */
        vector<SComponent> allSComponents;

        /** Vector with all possible R components of deterministic pda */
        vector<RComponent> allRComponents;

        /**
         * Divides transitions of rhdpda into internal, push and pop transitions.
         *
         * @return void
         */
        void divideTransitions();

        /**
         * Returns name of stack symbol of deterministic pda which is created from given S and R component and input symbol.
         *     This method also ensures that this stack symbol is already added into deterministic pda.
         *
         * @param s S component of stack symbol
         * @param r R component of stack symbol
         * @param input input symbol
         * @return name of stack symbol
         */
        string buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input);

        /**
         * Returns state and symbol pairs that are created by combining state from given set of states and given symbol.
         *
         * @param states
         * @param symbol
         * @return state and symbol pairs that are created by combining given states and symbol
         */
        set<StateSymbolPair> combineStatesWithSymbol(const set<State>& states, const Symbol& symbol);

        /**
         * Returns S component that contains pairs where both elements are same. These pairs are created from given set
         *     of state and symbol pairs.
         *
         * @param pairs set of state and symbol pairs from which to create identity pairs
         * @return pairs where both elements are same
         */
        const SComponent getSComponentWithPairsIdentity(const set<StateSymbolPair>& pairs);

        /**
         * Returns existing state from states map, if there is one with same name, or creates new one and adds it into
         *     states map and deterministic pda.
         *
         * @param s S component of state
         * @param r R component of state
         * @return state of deterministic pda
         */
        const State& getOrCreateState(const SComponent& s, const RComponent& r);

        /**
         * Runs some initializiation stuff before determinization algorithm.
         */
        void initDeterminization();

    public:

        /**
         * @param rhdpda real-time height-deterministic pushdown automaton given for determinization
         */
        RhdpdaDeterminizer3(PDA* rhdpda);

        /**
         * Runs determinization algorithm on rhdpda given in constructor.
         *
         * @return real-time deterministic pushdown automaton
         */
        Automaton* determinize();

};

}
}
#endif
