#ifndef RHDPDAUTILS_H_
#define RHDPDAUTILS_H_

#include <string>
#include <set>
#include <map>
#include <vector>

#include "automaton/PDA/PDA.h"
#include "automaton/Transition.h"
#include "automaton/exception/AutomatonException.h"
#include "alphabet/Symbol.h"

#include "../common/Utils.h"
#include "RhdpdaStructs.h"

using namespace std;
using namespace automaton;
using namespace alphabet;

namespace determinization {
namespace rhdpda {

/**
 * Library class with some methods that can be used by all rhdpda determinization algorithms.
 */
class RhdpdaUtils
{
    public:

        /** Symbol which represents bottom of stack */
        static const Symbol BOTTOM_OF_STACK_SYMBOL;

        /**
         * Returns string representation of given state and symbol pair.
         *
         * @param pair
         * @return name of pair
         */
        static string buildPairName(const StateSymbolPair& pair);

        /**
         * Returns name of state of deterministic pda which is created from given S and R component.
         *
         * @param s S component of state
         * @param r R component of state
         * @return name of state
         */
        static string buildStateName(const SComponent& s, const RComponent& r);

        /**
         * Returns existing state from states map, if there is one with same name, or creates new one and adds it into
         *     given states map and given pda.
         *
         * @param s S component of state
         * @param r R component of state
         * @param states map of existing states, where key is the name of state and value is data about this state
         * @param rdpda deterministic pda that contains given map of states
         * @return state of deterministic pda
         */
        static const State& getOrCreateState(const SComponent& s, const RComponent& r, map<string, StateData>& states,
            PDA& rdpda);

        /**
         * Returns S component that contains pairs where both elements are same. These pairs are created from given set
         *     of state and symbol pairs.
         *
         * @param pairs set of state and symbol pairs from which to create identity pairs
         * @return pairs where both elements are same
         */
        static const SComponent getSComponentWithPairsIdentity(const set<StateSymbolPair>& pairs);

        /**
         * Returns name of stack symbol of deterministic pda which is created from given S and R component and input symbol.
         *     This method also ensures that this stack symbol is already added into given deterministic pda.
         *
         * @param s S component of stack symbol
         * @param r R component of stack symbol
         * @param input input symbol
         * @param rdpda deterministic pda 
         * @return name of stack symbol
         */
        static string buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input, PDA& rdpda);
        
        /**
         * Returns all possible S components of deterministic pda.
         * 
         * @param pairs all possible pairs of state and symbol
         * @return all possible S components
         */
        static vector<SComponent> generateAllPossibleSComponents(set<StateSymbolPair>& pairs);

        /**
         * Returns all possible R components of deterministic pda.
         * 
         * @param pairs all possible pairs of state and symbol
         * @return all possible R components
         */
        static vector<RComponent> generateAllPossibleRComponents(set<StateSymbolPair>& pairs);

        /**
         * Returns state data about first state in given states map that isn't marked yet. Returns NULL if all states
         *     are already marked.
         *
         * @param states map of all states in deterministic pda
         * @return state data about first unmarked state
         */
        static StateData* getUnmarkedState(map<string, StateData>& states);

        /**
         * Returns state and symbol pairs that are created by combining state from given set of states and given symbol.
         *
         * @param states
         * @param symbol
         * @return state and symbol pairs that are created by combining given states and symbol
         */
        static set<StateSymbolPair> combineStatesWithSymbol(const set<State>& states, const Symbol& symbol);

        /**
         * Returns all possible state and symbol pairs created with states and symbols from given rhdpda.
         *
         * @param rhdpda real-time height-deterministic pda
         * @return all possible state and symbol pairs
         */
        static set<StateSymbolPair> buildAllStateSymbolPairs(PDA& rhdpda);
        
};

}
}
#endif
