#include "VpaDeterminizer.h"

namespace determinization {
namespace vpa {


void VpaDeterminizer::initDeterminization()
{
    this->states.clear();
    this->internalSymbols.clear();
    this->callSymbols.clear();
    this->returnSymbols.clear();
    Utils::copyInputAlphabet(*this->nvpa, *this->dvpa);
    VpaUtils::divideInputAlphabet(*this->nvpa, this->internalSymbols, this->callSymbols, this->returnSymbols);
    this->allSComponents = VpaUtils::generateAllSComponents(this->nvpa->getStates());
    this->allRComponents = VpaUtils::generateAllRComponents(this->nvpa->getStates());
    this->dvpa->addStackSymbol(VpaUtils::BOTTOM_OF_STACK_SYMBOL);
}


const State& VpaDeterminizer::getOrCreateState(const SComponent& s, const RComponent& r)
{
    return VpaUtils::getOrCreateState(s, r, this->states, *this->dvpa);
}


const SComponent VpaDeterminizer::getSComponentWithStatesIdentity(const set<State>& states)
{
    return VpaUtils::getSComponentWithStatesIdentity(states);
}


string VpaDeterminizer::buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input)
{
    return VpaUtils::buildStackSymbolName(s, r, input, *this->dvpa);
}


VpaDeterminizer::VpaDeterminizer(PDA* nvpa)
{
    this->nvpa = nvpa;
}


Automaton* VpaDeterminizer::determinize()
{
    this->dvpa = new PDA();
    this->initDeterminization();

    const SComponent& initialS = this->getSComponentWithStatesIdentity(this->nvpa->getStates());
    const RComponent& initialR(this->nvpa->getInitialStates());
    const State& initialState = this->getOrCreateState(initialS, initialR);
    this->dvpa->addInitialState(initialState);
    
    while (true) {
        StateData* unmarkedStateData = VpaUtils::getUnmarkedState(this->states);
        if (unmarkedStateData == NULL) {
            break;
        }
        
        // Internal symbols
        for (const auto& internalSymbol : internalSymbols) {
            SComponent sPrimed;
            SComponent& s = unmarkedStateData->s;
            for (const auto& statesPair : s.statesPairs) {
                const State& q = statesPair.state1;
                const State& qDoublePrimed = statesPair.state2;
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == qDoublePrimed && transition.getInput() == internalSymbol) {
                        const State& qPrimed = transition.getTo();
                        sPrimed.statesPairs.insert(StatesPair(q, qPrimed));
                    }
                }
            }
            RComponent rPrimed;
            RComponent& r = unmarkedStateData->r;
            for (const auto& q : r.states) {
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == q && transition.getInput() == internalSymbol) {
                        const State& qPrimed = transition.getTo();
                        rPrimed.states.insert(qPrimed);
                    }
                }
            }
            if (sPrimed.statesPairs.size() > 0 && rPrimed.states.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sPrimed, rPrimed);
                const TransitionPDA transition(fromState, internalSymbol, toState);
                this->dvpa->addTransition(transition);
            }
        }
        
        // Call symbols
        for (const auto& callSymbol : callSymbols) {
            RComponent rPrimed;
            RComponent& r = unmarkedStateData->r;
            for (const auto& q : r.states) {
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == q && transition.getInput() == callSymbol) {
                        const State& qPrimed = transition.getTo();
                        rPrimed.states.insert(qPrimed);
                    }
                }
            }
            if (rPrimed.states.size() > 0) {
                SComponent& s = unmarkedStateData->s;
                const SComponent& sPrimed = this->getSComponentWithStatesIdentity(this->nvpa->getStates());
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sPrimed, rPrimed);
                string stackSymbolName = this->buildStackSymbolName(s, r, callSymbol);
                list<Symbol> pop = {};
                list<Symbol> push = {stackSymbolName};
                const TransitionPDA transition(fromState, callSymbol, toState, pop, push);
                this->dvpa->addTransition(transition);
            }
        }
        
        // Return symbols
        for (const auto& returnSymbol : returnSymbols) {
            SComponent& s = unmarkedStateData->s;
            RComponent& r = unmarkedStateData->r;
            
            // Empty stack
            list<Symbol> bottomOfStack = {VpaUtils::BOTTOM_OF_STACK_SYMBOL};
            SComponent sPrimed;
            for (const auto& statesPair : s.statesPairs) {
                const State& q = statesPair.state1;
                const State& qDoublePrimed = statesPair.state2;
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == qDoublePrimed && transition.getInput() == returnSymbol &&
                            transition.getPop() == bottomOfStack) {
                        const State& qPrimed = transition.getTo();
                        sPrimed.statesPairs.insert(StatesPair(q, qPrimed));
                    }
                }
            }
            RComponent rPrimed;
            for (const auto& q : r.states) {
                for (const auto& transition : this->nvpa->getTransitions()) {
                    if (transition.getFrom() == q && transition.getInput() == returnSymbol &&
                            transition.getPop() == bottomOfStack) {
                        const State& qPrimed = transition.getTo();
                        rPrimed.states.insert(qPrimed);
                    }
                }
            }
            if (sPrimed.statesPairs.size() > 0 && rPrimed.states.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sPrimed, rPrimed);
                list<Symbol> pop = {VpaUtils::BOTTOM_OF_STACK_SYMBOL};
                list<Symbol> push = {};
                const TransitionPDA transition(fromState, returnSymbol, toState, pop, push);
                this->dvpa->addTransition(transition);
            }
            
            // Otherwise
            for (const auto& callSymbol :  callSymbols) {
                set<StatesPair> update;
                for (const auto& sPair : s.statesPairs) {
                    const State& q1 = sPair.state1;
                    const State& q2 = sPair.state2;
                    if (r.states.find(q2) != r.states.end()) {
                        for (const auto& transitionPush : this->nvpa->getTransitions()) {
                            for (const auto& transitionPop : this->nvpa->getTransitions()) {
                                const State& q = transitionPush.getFrom();
                                const State& qPrimed = transitionPop.getTo();
                                if (transitionPush.getInput() == callSymbol && transitionPush.getTo() == q1 &&
                                        transitionPop.getInput() == returnSymbol && transitionPop.getFrom() == q2 &&
                                        transitionPush.getPush() == transitionPop.getPop()) {
                                    update.insert(StatesPair(q, qPrimed));
                                }
                            }
                        }
                    }
                }
                if (update.size() == 0) {
                    continue;
                }
                for (const auto& sPrimed : this->allSComponents) {
                    for (const auto& rPrimed : this->allRComponents) {
                        RComponent rDoublePrimed;
                        for (const auto& updateItem : update) {
                            const State& q = updateItem.state1;
                            const State& qPrimed = updateItem.state2;
                            if (rPrimed.states.find(q) != rPrimed.states.end()) {
                                rDoublePrimed.states.insert(qPrimed);
                            }
                        }
                        SComponent sDoublePrimed;
                        for (const auto& updateItem : update) {
                            const State& q3 = updateItem.state1;
                            const State& qPrimed = updateItem.state2;
                            for (const auto& sPrimedPair : sPrimed.statesPairs) {
                                if (sPrimedPair.state2 == q3) {
                                    const State& q = sPrimedPair.state1;
                                    sDoublePrimed.statesPairs.insert(StatesPair(q, qPrimed));
                                }
                            }
                        }
                        if (sDoublePrimed.statesPairs.size() > 0 && rDoublePrimed.states.size() > 0) {
                            const State& fromState = unmarkedStateData->state;
                            const State& toState = this->getOrCreateState(sDoublePrimed, rDoublePrimed);
                            string stackSymbolName = this->buildStackSymbolName(sPrimed, rPrimed, callSymbol);
                            list<Symbol> pop = {Symbol(stackSymbolName)};
                            list<Symbol> push = {};
                            const TransitionPDA transition(fromState, returnSymbol, toState, pop, push);
                            this->dvpa->addTransition(transition);
                        }
                    }
                }
            }
        }
    }
    
    // Final states
    for (const auto& stateIter : states) {
        const StateData& stateData = stateIter.second;
        if (Utils::containFinalState(stateData.r.states, *this->nvpa)) {
            this->dvpa->addFinalState(stateData.state);
        }
    }
    
    return this->dvpa;
}


}
}
