#ifndef VPANUTILS_H_
#define VPANUTILS_H_

#include <math.h>
#include <string>
#include <set>
#include <map>
#include <vector>

#include "automaton/State.h"
#include "automaton/PDA/PDA.h"
#include "automaton/Transition.h"
#include "automaton/exception/AutomatonException.h"
#include "alphabet/Symbol.h"

#include "VpaStructs.h"

using namespace automaton;
using namespace alphabet;

namespace determinization {
namespace vpa {

/**
 * Library class with some methods that can be used by all vpa determinization algorithms.
 */
class VpaUtils
{
    public:

        /** Symbol which represents bottom of stack */
        static const Symbol BOTTOM_OF_STACK_SYMBOL;

        /**
         * Returns name of state of deterministic vpa which is created from given S and R component.
         *
         * @param s S component of state
         * @param r R component of state
         * @return name of state
         */
        static string buildStateName(const SComponent& s, const RComponent& r);

        /**
         * Returns existing state from states map, if there is one with same name, or creates new one and adds it into
         *     given states map and given vpa.
         *
         * @param s S component of state
         * @param r R component of state
         * @param states map of existing states, where key is the name of state and value is data about this state
         * @param dvpa deterministic vpa that contains given map of states
         * @return state of deterministic vpa
         */
        static const State& getOrCreateState(const SComponent& s, const RComponent& r, map<string, StateData>& states,
            PDA& dvpa);

        /**
         * Returns S component that contains pairs where states are same.
         *
         * @param states set of states from which to create pairs
         * @return pairs where states are same
         */
        static const SComponent getSComponentWithStatesIdentity(const set<State>& states);

        /**
         * Divides input alphabet of given pda into internal, call and return symbols.
         *
         * @param vpa vpa with input alphabet for dividing
         * @param internalSymbols output parameter with internal symbols
         * @param callSymbols output parameter with call symbols
         * @param returnSymbols output parameter with return symbols
         * @return void
         */
        static void divideInputAlphabet(PDA& vpa, set<Symbol>& internalSymbols, set<Symbol>& callSymbols,
            set<Symbol>& returnSymbols);

        /**
         * Returns name of stack symbol of deterministic vpa which is created from given S and R component and input symbol.
         *     This method also ensures that this stack symbol is already added into given deterministic vpa.
         *
         * @param s S component of stack symbol
         * @param r R component of stack symbol
         * @param input input symbol
         * @param vpa deterministic vpa 
         * @return name of stack symbol
         */
        static string buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input, PDA& dvpa);

        /**
         * Returns all possible S components of deterministic vpa.
         * 
         * @param states all states of nondeterministic vpa
         * @return all possible S components
         */
        static vector<SComponent> generateAllSComponents(const set<State>& states);

        /**
         * Returns all possible R components of deterministic vpa.
         * 
         * @param states all states of nondeterministic vpa
         * @return all possible R components
         */
        static vector<RComponent> generateAllRComponents(const set<State>& states);

        /**
         * Returns state data about first state in given states map that isn't marked yet. Returns NULL if all states
         *     are already marked.
         *
         * @param states map of all states in vpa
         * @return state data about first unmarked state
         */
        static StateData* getUnmarkedState(map<string, StateData>& states);

};


}
}
#endif
