#include <iostream>
#include <getopt.h>
#include <string.h>
#include <ctype.h>

#include "automaton/Automaton.h"
#include "automaton/FSM/NFA.h"
#include "automaton/FSM/DFA.h"
//#include "automaton/PDA/PDA.h"
#include "factory/DataFactory.hpp"
#include "exception/AlibException.h"

#include "determinize/nfa/NFADeterminizer.h"
//#include "idpda/IdpdaDeterminizer.h"
//#include "vpa/VpaDeterminizer.h"
//#include "vpa/VpaDeterminizer2.h"
//#include "vpa/VpaDeterminizer3.h"
//#include "rhdpda/RhdpdaDeterminizer.h"
//#include "rhdpda/RhdpdaDeterminizer2.h"
//#include "rhdpda/RhdpdaDeterminizer3.h"
//#include "rhdpda/RhdpdaDeterminizer4.h"

#define VERSION "0.0.1"

#define TYPE_FSM "fsm"
#define TYPE_IDPDA "idpda"
#define TYPE_VPA "vpa"
#define TYPE_RHDPDA "rhdpda"

#define VERSION_1 "1"
#define VERSION_2 "2"
#define VERSION_3 "3"
#define VERSION_4 "4"


/**
 * Prints help to standard output.
 */
void printHelp() {
	std::cout << "adeterminize " << VERSION << std::endl;
	std::cout << "Determinize various types of automaton." << std::endl;
	std::cout << "Usage: adeterminize -t TYPE [SWITCH...]" << std::endl;
	std::cout << "Possible arguments:" << std::endl;
	std::cout << "  -t, --type=TYPE \t Specifies type of input automaton, possible values are 'FSM' for final-state machine, 'IDPDA' for input-driven pushdown automaton, 'VPA' for visible pushdown automaton and 'RHDPDA' for real-time height deterministic pushdown automaton." << std::endl;
	std::cout << "  -a, --algorithm=VERSION_OF_ALGORITHM \t Specifies version of algorithm. This argument works only with VPA or RHDPDA type. VPA determinization has 3 versions (use numbers 1, 2 and 3) and RHDPDA determinization has 4 versions (use numbers 1, 2, 3 and 4). Default value is always number 1." << std::endl;
	std::cout << "  -h, --help \t Displays this help message." << std::endl;
	std::cout << std::endl;
}


/**
 * Converts given type to lower case.
 *
 * @param type
 * @return type in lower case
 */
std::string getLowerCaseType(std::string type) {
	std::string lowerCaseType;
	for (const auto& c : type) {
		lowerCaseType.append(1, tolower(c));
	}
	return lowerCaseType;
}


int main(int argc, char** argv) {
	static struct option longOptions[] = {
	  {"type", required_argument, NULL, 't'},
	  {"algorithm", required_argument, NULL, 'a'},
	  {"version", no_argument, NULL, 'v'},
	  {"help", no_argument, NULL, 'h'},
	  {0, 0, 0, 0}
	};

	int longIndex = 0;
	int opt = 0;
	std::string type;
	std::string version = "1";

	while ((opt = getopt_long(argc, argv, "t:a:vh", longOptions, &longIndex)) != -1) {
		switch(opt) {
			case 't':
				type.assign(optarg, strlen(optarg));
				break;
			case 'a':
				version.assign(optarg, strlen(optarg));
				break;
			case 'v':
			case 'h':
			default:
				printHelp();
				return 0;
		}
	}

	type = getLowerCaseType(type);

	std::list<sax::Token> tokens;
	sax::SaxParseInterface::parseStdin(tokens);

	try {
		if (type == TYPE_FSM) {
			std::string xmlMark = tokens.front( ).getData( );
			if(xmlMark == "NFA") {
				automaton::NFA nfa = alib::DataFactory::fromTokens<automaton::NFA>(tokens);
				automaton::DFA dfa = determinize::NFADeterminizer::determinize(nfa);
				alib::DataFactory::toStdout(dfa);
				return 0;
			} else if(xmlMark == "DFA") {
				automaton::DFA automaton = alib::DataFactory::fromTokens<automaton::DFA>(tokens);
				alib::DataFactory::toStdout(automaton);
				return 0;
			} else {
				throw exception::AlibException("Unsupported formalism");
			}
		}

/*		} else if (type == TYPE_IDPDA) {
			return new idpda::IdpdaDeterminizer((PDA*) automaton);

		} else if (type == TYPE_VPA) {
			return getVpaDeterminizer(automaton, version);

		} else if (type == TYPE_RHDPDA) {
			return getRhdpdaDeterminizer(automaton, version);
		}*/

		printHelp(); // should not be reached
		return 2;
	} catch (const exception::AlibException& e) {
		alib::DataFactory::toStdout(e);
		return 1;
	} catch (...) {
		return 127;
	}
}
