/*
 * EpsilonClosure.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "EpsilonClosure.h"

#include <exception/AlibException.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

#include "../regexp/RegExpEpsilon.h"

namespace automaton {

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::EpsilonNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getEpsilonTransitionsFromState( p ) )
			for (const auto & to : transition.second )
				if( visited [ to ] == false )
					queue.push( to );
	}

	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::NFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::DFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::ExtendedNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( regexp::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::CompactNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( transition.first.second.getContent().size() != 0 )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure(const Automaton& automaton, const automaton::State& q) {
	std::pair<automaton::State, std::set<automaton::State>> out = std::make_pair(q, std::set<automaton::State> {});
	automaton.getData().Accept((void*) &out, EpsilonClosure::EPSILON_CLOSURE);
	return out.second;
}

void EpsilonClosure::Visit(void*, const UnknownAutomaton&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void EpsilonClosure::Visit(void* data, const EpsilonNFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = this->epsilonClosure(automaton, out.first);
}

void EpsilonClosure::Visit(void* data, const NFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = this->epsilonClosure(automaton, out.first);
}

void EpsilonClosure::Visit(void* data, const DFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = this->epsilonClosure(automaton, out.first);
}

void EpsilonClosure::Visit(void* data, const ExtendedNFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = this->epsilonClosure(automaton, out.first);
}

void EpsilonClosure::Visit(void* data, const CompactNFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = this->epsilonClosure(automaton, out.first);
}

void EpsilonClosure::Visit(void*, const DPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void EpsilonClosure::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void EpsilonClosure::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void EpsilonClosure::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void EpsilonClosure::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

const EpsilonClosure EpsilonClosure::EPSILON_CLOSURE;

}

