/*
 * BrzozowskiAlgebraic.cpp
 *
 *  Created on: 11. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "BrzozowskiAlgebraic.h"
#include "alphabet/Symbol.h"
#include "automaton/FSM/DFA.h"
#include "automaton/FSM/NFA.h"
#include "automaton/FSM/EpsilonNFA.h"

namespace fa2re {

template<>
regexp::UnboundedRegExp BrzozowskiAlgebraic::convert( const automaton::EpsilonNFA & automaton ) {
	equations::RightRegularEquationSolver solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( alphabet::Symbol( alphabet::LabeledSymbol( q.getName( ) ) ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( isInSet( q, automaton.getFinalStates( ) ) )
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( q.getName( ) ) ), regexp::UnboundedRegExpEpsilon { } );
	}

	for( const auto & p : automaton.getSymbolTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( q.getName() ) ), regexp::UnboundedRegExpSymbol { p.first.second } );
		}
	}

	for( const auto & p : automaton.getEpsilonTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( q.getName() ) ), regexp::UnboundedRegExpEpsilon { } );
		}
	}

	regexp::UnboundedRegExpAlternation alternation;
	for(const auto& initialSymbol : automaton.getInitialStates() ) {
		// set symbol for which the solver will solve equation system
		alphabet::Symbol tmp( alphabet::LabeledSymbol ( initialSymbol.getName( ) ) );

		alternation.appendElement( solver.solve( tmp ).getRegExp() );
	}

	return regexp::UnboundedRegExp { alternation };
}

template<>
regexp::UnboundedRegExp BrzozowskiAlgebraic::convert( const automaton::NFA & automaton ) {
	equations::RightRegularEquationSolver solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( alphabet::Symbol( alphabet::LabeledSymbol( q.getName( ) ) ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( isInSet( q, automaton.getFinalStates( ) ) )
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( q.getName( ) ) ), regexp::UnboundedRegExpEpsilon { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( q.getName() ) ), regexp::UnboundedRegExpSymbol { p.first.second } );
		}
	}

	regexp::UnboundedRegExpAlternation alternation;
	for(const auto& initialSymbol : automaton.getInitialStates() ) {
		// set symbol for which the solver will solve equation system
		alphabet::Symbol tmp( alphabet::LabeledSymbol ( initialSymbol.getName( ) ) );

		alternation.appendElement( solver.solve( tmp ).getRegExp() );
	}

	return regexp::UnboundedRegExp { alternation };
}

template<>
regexp::UnboundedRegExp BrzozowskiAlgebraic::convert( const automaton::DFA & automaton ) {
	equations::RightRegularEquationSolver solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( alphabet::Symbol( alphabet::LabeledSymbol( q.getName( ) ) ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( isInSet( q, automaton.getFinalStates( ) ) )
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( q.getName( ) ) ), regexp::UnboundedRegExpEpsilon { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( p.second.getName() ) ), regexp::UnboundedRegExpSymbol { p.first.second } );
	}

	return solver.solve( alphabet::Symbol( alphabet::LabeledSymbol (automaton.getInitialState().getName() ) ) );
}

} /* namespace fa2re */
