#include "FAtoRRGConverter.h"

#include <label/StringLabel.h>
#include <alphabet/LabeledSymbol.h>

#include <exception/AlibException.h>

namespace fa2rg
{

grammar::RightRG FAtoRRGConverter::convert(const automaton::NFA& automaton)
{
	std::map<automaton::State, alphabet::Symbol> nonterminalMap;

	const automaton::State& initState = *automaton.getInitialStates().begin(); //FIXME what about other initial states
	const alphabet::Symbol initSymbol(alphabet::LabeledSymbol(initState.getName()));

	grammar::RightRG grammar(initSymbol);
	nonterminalMap.insert(std::make_pair(initState, initSymbol));

	grammar.setTerminalAlphabet(automaton.getInputAlphabet());

	for(const auto& state : automaton.getStates())
	{
		if(state == initState)
			continue;

		alphabet::Symbol nt = alphabet::createUniqueSymbol(alphabet::Symbol(alphabet::LabeledSymbol(state.getName())), grammar.getTerminalAlphabet(), grammar.getNonterminalAlphabet());
		grammar.addNonterminalSymbol(nt);
		nonterminalMap.insert(std::make_pair(state, nt));
	}

	// step 2 - create set of P in G
	for(const auto& transition : automaton.getTransitions())
	{
		const automaton::State& from = transition.first.first;
		const alphabet::Symbol& input = transition.first.second;
		for(const auto& to : transition.second)
		{
			grammar.addRule(nonterminalMap.find(from)->second, std::make_pair(input, nonterminalMap.find(to)->second)); // 2a
			if(automaton.getFinalStates().count(to)) // 2b
				grammar.addRule(nonterminalMap.find(from)->second, input);
		}
	}

	// step 3 - set start symbol of G
	grammar.setInitialSymbol(nonterminalMap.find(*automaton.getInitialStates().begin())->second);

	// step 4
	for(const auto & initial : automaton.getInitialStates()) {
		if(automaton.getFinalStates().count(initial) > 0)
			grammar.setGeneratesEpsilon(true); // okay this feature makes algorithm mismatch mathematical but simplifies the code actually :))
	}

	return grammar;
}

grammar::RightRG FAtoRRGConverter::convert(const automaton::DFA& automaton)
{
	std::map<automaton::State, alphabet::Symbol> nonterminalMap;

	const automaton::State& initState = automaton.getInitialState();
	const alphabet::Symbol initSymbol(alphabet::LabeledSymbol(initState.getName()));

	grammar::RightRG grammar(initSymbol);
	nonterminalMap.insert(std::make_pair(initState, initSymbol));

	grammar.setTerminalAlphabet(automaton.getInputAlphabet());

	for(const auto& state : automaton.getStates())
	{
		if(state == initState)
			continue;

		alphabet::Symbol nt = alphabet::createUniqueSymbol(alphabet::Symbol(alphabet::LabeledSymbol(state.getName())), grammar.getTerminalAlphabet(), grammar.getNonterminalAlphabet());
		grammar.addNonterminalSymbol(nt);
		nonterminalMap.insert(std::make_pair(state, nt));
	}

	// step 2 - create set of P in G
	for(const auto& transition : automaton.getTransitions())
	{
		const automaton::State& from = transition.first.first;
		const alphabet::Symbol& input = transition.first.second;
		const automaton::State& to = transition.second;

		grammar.addRule(nonterminalMap.find(from)->second, std::make_pair(input, nonterminalMap.find(to)->second)); // 2a
		if(automaton.getFinalStates().count(to)) // 2b
			grammar.addRule(nonterminalMap.find(from)->second, input);
	}

	// step 3 - set start symbol of G
	grammar.setInitialSymbol(nonterminalMap.find(automaton.getInitialState())->second);

	// step 4
	if(automaton.getFinalStates().count(automaton.getInitialState()))
		grammar.setGeneratesEpsilon(true); // okay this feature breaks algorithm but simplifies the code actually :))

	return grammar;
}

grammar::Grammar FAtoRRGConverter::convert(const automaton::Automaton& automaton) {
	grammar::Grammar* out = NULL;
	automaton.getData().Accept((void*) &out, FAtoRRGConverter::FA_TO_RRG_CONVERTER);
	grammar::Grammar res = std::move(*out);
	delete out;
	return res;
}

void FAtoRRGConverter::Visit(void*, const automaton::UnknownAutomaton&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void*, const automaton::EpsilonNFA& ) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void* data, const automaton::NFA& automaton) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->convert(automaton));
}

void FAtoRRGConverter::Visit(void* data, const automaton::DFA& automaton) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->convert(automaton));
}

void FAtoRRGConverter::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void*, const automaton::DPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void*, const automaton::SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void*, const automaton::NPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void*, const automaton::SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void FAtoRRGConverter::Visit(void*, const automaton::OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

const FAtoRRGConverter FAtoRRGConverter::FA_TO_RRG_CONVERTER;

} /* namespace fa2rg */
