/*
 * RegularEquationSolver.cpp
 *
 *  Created on: 4. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "RegularEquationSolver.h"

#include "regexp/unbounded/UnboundedRegExpElements.h"

namespace equations {

regexp::UnboundedRegExp RegularEquationSolver::solve( const alphabet::Symbol & solveFor ) {
	if( ! isInSet( solveFor, m_symbols ) )
		throw exception::AlibException( "Trying to solve for symbol not present in equation system" );

	/*
	 * Firstly, organize states by depth so we can output better looking
	 * expressions. We need to solve equation system for automaton initial state,
	 * so lets start with the deepest ones and walk towards initial one.
	 */
	symbolsByDepth( solveFor );

	return eliminate( );
}

void RegularEquationSolver::addSymbol( const alphabet::Symbol & symb ) {
	m_symbols.insert( symb );

	for( const auto & s1 : m_symbols )
	{
		m_eqTransition[ symb ][ s1 ] = regexp::UnboundedRegExpAlternation { };
		if( s1 != symb )
			m_eqTransition[ s1 ][ symb ] = regexp::UnboundedRegExpAlternation { };
	}

	m_eqFinal[ symb ] = regexp::UnboundedRegExpAlternation { };
}

void RegularEquationSolver::addEquation( const alphabet::Symbol & from, const alphabet::Symbol & to, const regexp::UnboundedRegExpElement& eq ) {
	if( ! isInSet( from, m_symbols ) )
		throw exception::AlibException( "Symbol '" + (std::string) from + "' not in equation system." );

	if( ! isInSet( to, m_symbols ) )
		throw exception::AlibException( "Symbol '" + (std::string) to + "' not in equation system." );

	m_eqTransition[ from ][ to ].appendElement( eq );
}

void RegularEquationSolver::addEquation( const alphabet::Symbol & from, const regexp::UnboundedRegExpElement& eq ) {
	if( ! isInSet( from, m_symbols ) )
		throw exception::AlibException( "Symbol '" + (std::string) from + "' not in equation system." );

	m_eqFinal[ from ].appendElement( eq );
}

void RegularEquationSolver::symbolsByDepth( const alphabet::Symbol & solveFor ) {
	std::map<alphabet::Symbol, bool> visited;
	std::queue<alphabet::Symbol> queue;

	for( const auto & symbol : m_symbols )
		visited[ symbol ] = false;

	visited[ solveFor ] = true;
	queue.push( solveFor );

	while( ! queue.empty( ) )
	{
		alphabet::Symbol s = queue.front( );
		queue.pop( );

		m_symbolsByDepth.push_back( s );

		auto map_s = m_eqTransition.find( s );
		for( auto it = map_s->second.begin( ); it != map_s->second.end( ); it ++ )
		{
			if( it->second.getElements().size() > 0 && visited[ it->first ] == false ) {
				visited[ it->first ] = true;
				queue.push( it->first );
			}
		}
	}
}

} /* namespace equations */
