/*
 * UnknownAutomaton.cpp
 *
 *  Created on: Oct 12, 2013
 *      Author: Martin Zak
 */

#include "UnknownAutomaton.h"
#include "../std/set.hpp"
#include <sstream>
#include "../std/pointer.hpp"

#include "AutomatonException.h"

namespace automaton {

UnknownAutomaton::UnknownAutomaton() :
		blankSymbol(NULL) {

}

UnknownAutomaton::~UnknownAutomaton() {
}

AutomatonBase* UnknownAutomaton::clone() const {
	return new UnknownAutomaton(*this);
}

AutomatonBase* UnknownAutomaton::plunder() && {
	return new UnknownAutomaton(std::move(*this));
}

bool UnknownAutomaton::addState(const State& state) {
	return states.insert(state).second;
}

void UnknownAutomaton::setStates(const std::set<State>& states) {
	this->states = states; 
}

bool UnknownAutomaton::removeState(const State& state) {
	return states.erase(state);
}

const std::set<State>& UnknownAutomaton::getStates() const {
	return states;
}


bool UnknownAutomaton::addInitialState(const State& state) {
	return initialStates.insert(state).second;
}

void UnknownAutomaton::setInitialStates(const std::set<State>& states) {
	initialStates = states;
}

bool UnknownAutomaton::removeInitialState(const State& state) {
	return initialStates.erase(state);
}

const std::set<State>& UnknownAutomaton::getInitialStates() const {
	return initialStates;
}


bool UnknownAutomaton::addFinalState(const State& state) {
	return finalStates.insert(state).second;
}

void UnknownAutomaton::setFinalStates(const std::set<State>& states) {
	finalStates = states;
}

bool UnknownAutomaton::removeFinalState(const State& state) {
	return finalStates.erase(state);
}

const std::set<State>& UnknownAutomaton::getFinalStates() const {
	return finalStates;
}


bool UnknownAutomaton::addInputSymbol(const alphabet::Symbol& symbol) {
	return inputAlphabet.insert(symbol).second;
}

const std::set<alphabet::Symbol>& UnknownAutomaton::getInputSymbols() const {
	return this->inputAlphabet;
}

void UnknownAutomaton::setInputSymbols(const std::set<alphabet::Symbol>& symbols) {
	this->inputAlphabet = symbols;
}

bool UnknownAutomaton::removeInputSymbol(const alphabet::Symbol& symbol) {
	return inputAlphabet.erase(symbol);
}


bool UnknownAutomaton::addStackSymbol(const alphabet::Symbol& symbol) {
	return stackAlphabet.insert(symbol).second;
}

void UnknownAutomaton::setStackSymbols(const std::set<alphabet::Symbol>& symbols) {
	stackAlphabet = symbols;
}

bool UnknownAutomaton::removeStackSymbol(const alphabet::Symbol& symbol) {
	return stackAlphabet.erase(symbol);
}

const std::set<alphabet::Symbol>& UnknownAutomaton::getStackSymbols() const {
	return stackAlphabet;
}


bool UnknownAutomaton::addInitialSymbol(const alphabet::Symbol& symbol) {
	return stackAlphabet.insert(symbol).second;
}

void UnknownAutomaton::setInitialSymbols(const std::set<alphabet::Symbol>& symbols) {
	this->initialSymbols = symbols;
}

bool UnknownAutomaton::removeInitialSymbol(const alphabet::Symbol& symbol) {
	return stackAlphabet.erase(symbol);
}

const std::set<alphabet::Symbol>& UnknownAutomaton::getInitialSymbols() const {
	return initialSymbols;
}


bool UnknownAutomaton::addTapeSymbol(const alphabet::Symbol& symbol) {
	return tapeAlphabet.insert(symbol).second;
}

void UnknownAutomaton::setTapeSymbols(const std::set<alphabet::Symbol>& symbols) {
	tapeAlphabet = symbols;
}

bool UnknownAutomaton::removeTapeSymbol(const alphabet::Symbol& symbol) {
	return tapeAlphabet.erase(symbol);
}

const std::set<alphabet::Symbol>& UnknownAutomaton::getTapeSymbols() const {
	return tapeAlphabet;
}


void UnknownAutomaton::setBlankSymbol(const alphabet::Symbol& symbol) {
	delete blankSymbol;
	blankSymbol = new alphabet::Symbol(symbol);
}

bool UnknownAutomaton::hasBlankSymbol() const {
	return blankSymbol != NULL;
}

void UnknownAutomaton::clearBlankSymbol() {
	delete blankSymbol;
	blankSymbol = NULL;
}

const alphabet::Symbol& UnknownAutomaton::getBlankSymbol() const {
	return *blankSymbol;
}


bool UnknownAutomaton::addTransition(const UnknownTransition& transition) {
	return transitions.insert(transition).second;
}

bool UnknownAutomaton::removeTransition(const UnknownTransition& transition) {
	return transitions.erase(transition);
}

const std::set<UnknownTransition>& UnknownAutomaton::getTransitions() const {
	return transitions;
}


bool UnknownAutomaton::operator==(const ObjectBase& other) const {
	return other == *this;
}

bool UnknownAutomaton::operator<(const ObjectBase& other) const {
	return other > *this;
}

bool UnknownAutomaton::operator>(const ObjectBase& other) const {
	return other < *this;
}

bool UnknownAutomaton::operator==(const UnknownAutomaton& other) const {
	std::pointer<alphabet::Symbol> blankSymbolPointer(blankSymbol);
	std::pointer<alphabet::Symbol> otherBlankSymbolPointer(other.blankSymbol);

	return states == other.states
		&& inputAlphabet == other.inputAlphabet
		&& initialStates == other.initialStates
		&& finalStates == other.finalStates
		&& stackAlphabet == other.stackAlphabet
		&& initialSymbols == other.initialSymbols
		&& tapeAlphabet == other.tapeAlphabet
		&& blankSymbolPointer == otherBlankSymbolPointer
		&& transitions == other.transitions;
}

bool UnknownAutomaton::operator<(const UnknownAutomaton& other) const {
	std::pointer<alphabet::Symbol> blankSymbolPointer(blankSymbol);
	std::pointer<alphabet::Symbol> otherBlankSymbolPointer(other.blankSymbol);

	return std::tie(states, inputAlphabet, initialStates, finalStates, stackAlphabet, initialSymbols, tapeAlphabet, blankSymbolPointer, transitions) < std::tie(other.states, other.inputAlphabet, other.initialStates, other.finalStates, other.stackAlphabet, other.initialSymbols, other.tapeAlphabet, otherBlankSymbolPointer, other.transitions);
}

void UnknownAutomaton::operator>>(std::ostream& out) const {
	std::pointer<alphabet::Symbol> blankSymbolPointer(blankSymbol);

	out << "(UnknownAutomaton"
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialStates = " << initialStates
		<< " finalStates = " << finalStates
		<< " stackAlphabet = " << stackAlphabet
		<< " initialSymbols = " << initialSymbols
		<< " tapeAlphabet = " << tapeAlphabet
		<< " blankSymbol = " << blankSymbolPointer
		<< " transitions = " << transitions
		<< ")";
}

UnknownAutomaton::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace automaton */

