/*
 * AutomatonFactory.h
 *
 *  Created on: Apr 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef AUTOMATON_CONVERTOR_H_
#define AUTOMATON_FONVERTOR_H_

#include "../automaton/Automaton.h"
#include "../automaton/UnknownAutomaton.h"
#include "../automaton/FSM/EpsilonNFA.h"
#include "../automaton/PDA/NPDA.h"
#include "../automaton/TM/OneTapeDTM.h"

namespace automaton {

/**
 * Builds specific automatons from UnknownAutomaton. It converts an UnknownAutomaton to the most general automaton in the group - EpsilonNFA for FSA, ...
 */
class AutomatonConvertor {
public:
	/**
	 * Tries to create specific automaton from UnknownAutomaton.
	 * @param automaton source UnknownAutomaton
	 * @return pointer to new automaton
	 * @throws AutomatonException when specific automaton cannot
	 * be created form UnknownAutomaton (e.g. automaton type is not recognized)
	 */
	static Automaton buildAutomaton(const UnknownAutomaton& automaton);

	/**
	 * Checks that FSM can be build from UnknownAutomaton. That means UnknownAutomaton
	 * doesn't contain elements that are not present in the FSM.
	 * @param automaton source UnknownAutomaton
	 * @return true when FSM can be build from UnknownAutomaton, false otherwise
	 */
	static bool isEpsilonNFA(const UnknownAutomaton& automaton);

	/**
	 * Tries to build FSM from the UnknownAutomaton. Ignores elements that
	 * are not required for the FSM.
	 * @param automaton source UnknownAutomaton
	 * @return EpsilonNFA
	 */
	static EpsilonNFA buildEpsilonNFA(const UnknownAutomaton& automaton);

	/**
	 * Checks that PDA can be build from UnknownAutomaton. That means UnknownAutomaton
	 * doesn't contain elements that are not present in the PDA.
	 * @param automaton source UnknownAutomaton
	 * @return true when PDA can be build from UnknownAutomaton, false otherwise
	 */
	static bool isNPDA(const UnknownAutomaton& automaton);

	/**
	 * Tries to build PDA from the UnknownAutomaton. Ignores elements that
	 * are not required for the PDA.
	 * @param automaton source UnknownAutomaton
	 * @return PDA
	 */
	static NPDA buildNPDA(const UnknownAutomaton& automaton);

	/**
	 * Checks that TM can be build from UnknownAutomaton. That means UnknownAutomaton
	 * doesn't contain elements that are not present in the TM.
	 * @param automaton source UnknownAutomaton
	 * @return true when TM can be build from UnknownAutomaton, false otherwise
	 */
	static bool isOneTapeDTM(const UnknownAutomaton& automaton);

	/**
	 * Tries to build TM from the UnknownAutomaton. Ignores elements that
	 * are not required for the TM.
	 * @param automaton source UnknownAutomaton
	 * @return TM
	 */
	static OneTapeDTM buildOneTapeDTM(const UnknownAutomaton& automaton);

protected:
	/**
	 * Build common part (states, input alphabet, initial states, final states)
	 * of the automaton from UnknownAutomaton.
	 * @param automaton automaton to build (destination)
	 * @param unknownAutomaton source automaton
	 */
	static void buildCommon(Automaton& automaton, const UnknownAutomaton& unknownAutomaton);

	/**
	 * Checks that transition contains only elements that are valid for FSM.
	 * @param transition UnknownTransition to check
	 * @return true when transition is FSM transition
	 */
	static bool isEpsilonNFATransition(const UnknownTransition& transition);

	/**
	 * Checks that transition contains only elements that are valid for PDA.
	 * @param transition UnknownTransition to check
	 * @return true when transition is PDA transition
	 */
	static bool isNPDATransition(const UnknownTransition& transition);

	/**
	 * Checks that transition contains only elements that are valid for TM.
	 * @param transition UnknownTransition to check
	 * @return true when transition is TM transition
	 */
	static bool isOneTapeDTMTransition(const UnknownTransition& transition);
};

} /* namespace automaton */

#endif /* AUTOMATON_CONVERTOR_H_ */

