/*
 * GNF.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "GNF.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>

#include "../../alphabet/Symbol.h"

namespace grammar {

GNF::GNF(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {

}

GNF::GNF(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* GNF::clone() const {
	return new GNF(*this);
}

GrammarBase* GNF::plunder() && {
	return new GNF(std::move(*this));
}

bool GNF::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol> >> >& rule : rules) {
		for(const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol> >& rhs : rule.second)
			if(rhs.first == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool GNF::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol> >> >& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol> >& rhs : rule.second)
			if(std::find(rhs.second.begin(), rhs.second.end(), symbol) != rhs.second.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool GNF::addRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol> >& rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	if(!terminalAlphabet.count(rightHandSide.first))
		throw GrammarException("Rule must rewrite to terminal symbol");

	for(const alphabet::Symbol& rhsNTs : rightHandSide.second)
		if(nonterminalAlphabet.find(rhsNTs) == nonterminalAlphabet.end())
			throw GrammarException("Symbol \"" + (std::string) rhsNTs + "\" is not a nonterminal symbol");

	return rules[leftHandSide].insert(rightHandSide).second;
}

const std::map<alphabet::Symbol, std::set<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol> >> > & GNF::getRules() const {
	return rules;
}

bool GNF::removeRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol> >& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

bool GNF::addRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		if(leftHandSide != initialSymbol) throw exception::AlibException("Illegal left hand side of epsilon rule");
		bool res = getGeneratesEpsilon();
		setGeneratesEpsilon(true);
		return !res;
	} else {
		alphabet::Symbol first = rightHandSide[0];
		std::vector<alphabet::Symbol> rest(rightHandSide.begin() + 1, rightHandSide.end());
		return addRule(leftHandSide, std::make_pair(first, rest));
	}
}

std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> GNF::getRawRules() const {
	std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> res;
	for(const auto& rule : getRules()) {
		for(const auto& rhs : rule.second) {
			std::vector<alphabet::Symbol> tmp;
			tmp.push_back(rhs.first);
			tmp.insert(tmp.end(), rhs.second.begin(), rhs.second.end());
			res[rule.first].insert(tmp);
		}
	}
	if(getGeneratesEpsilon()) {
		res[getInitialSymbol()].insert(std::vector<alphabet::Symbol> {});
	}
	return res;
}

bool GNF::removeRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		if(leftHandSide != initialSymbol) throw exception::AlibException("Illegal left hand side of epsilon rule");
		bool res = getGeneratesEpsilon();
		setGeneratesEpsilon(false);
		return res;
	} else {
		alphabet::Symbol first = rightHandSide[0];
		std::vector<alphabet::Symbol> rest(rightHandSide.begin() + 1, rightHandSide.end());
		return removeRule(leftHandSide, std::make_pair(first, rest));
	}
}

void GNF::setGeneratesEpsilon(bool genEps) {
	generatesEpsilon = genEps;
}

bool GNF::getGeneratesEpsilon() const {
	return generatesEpsilon;
}

bool GNF::operator==(const ObjectBase& other) const {
	return other == *this;
}

bool GNF::operator<(const ObjectBase& other) const {
	return other > *this;
}

bool GNF::operator>(const ObjectBase& other) const {
	return other < *this;
}

bool GNF::operator==(const GNF& other) const {
	return this->nonterminalAlphabet == other.nonterminalAlphabet && this->terminalAlphabet == other.terminalAlphabet && this->initialSymbol == other.initialSymbol && this->rules == other.rules;
}

bool GNF::operator<(const GNF& other) const {
	return std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules) < std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);
}

void GNF::operator>>(std::ostream& out) const {
	out << "(GNF"
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

GNF::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace grammar */
