/*
 * GrammarFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "GrammarFromStringParser.h"

#include "../exception/AlibException.h"

#include "UnknownGrammar.h"

namespace grammar {

GrammarFromStringParser::GrammarFromStringParser(std::istream& input) : m_GrammarLexer(input), m_SymbolParser(input) {

}

Grammar GrammarFromStringParser::parse() {
	return parse(std::set<FEATURES>({FEATURES::GRAMMAR}));
}

Grammar GrammarFromStringParser::parse(const std::set<FEATURES>& features) {
	if(!features.count(FEATURES::GRAMMAR)) throw exception::AlibException();
	return Grammar(parseUnknownGrammar());
}

bool GrammarFromStringParser::first() {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next().token();
	if(token.type == GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		return true;
	} else {
		return false;
	}
}

bool GrammarFromStringParser::next() {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next().token();
	if(token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SEPARATOR || token.type == GrammarFromStringLexer::TokenType::SET_BEGIN || token.type == GrammarFromStringLexer::TokenType::SET_END || token.type == GrammarFromStringLexer::TokenType::TUPLE_BEGIN || token.type == GrammarFromStringLexer::TokenType::TUPLE_END || token.type == GrammarFromStringLexer::TokenType::EPSILON || token.type == GrammarFromStringLexer::TokenType::MAPS_TO) {
		return true;
	} else {
		return false;
	}
}

bool GrammarFromStringParser::last() {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next().token();
	if(token.type == GrammarFromStringLexer::TokenType::TEOF) {
		return true;
	} else {
		return false;
	}
}

UnknownGrammar GrammarFromStringParser::parseUnknownGrammar() {
	UnknownGrammar res;

	GrammarFromStringLexer::Token token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		throw exception::AlibException();
	}
	next();

	std::set<alphabet::Symbol> nonterminals = parseSet();
	res.setNonterminalAlphabet(nonterminals);

	token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException();
	}
	next();

	std::set<alphabet::Symbol> terminals = parseSet();
	res.setTerminalAlphabet(terminals);

	token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException();
	}
	next();

	parseRules(res);

	token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException();
	}
	next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();

	alphabet::Symbol initialSymbol = m_SymbolParser.parse();
	res.setInitialSymbol(initialSymbol);

	next();
	token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_END) {
		throw exception::AlibException();
	}

	return  res;
}

std::set<alphabet::Symbol> GrammarFromStringParser::parseSet() {
	std::set<alphabet::Symbol> res;

	GrammarFromStringLexer::Token token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::AlibException();
	}
	next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();

	token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) while(true) {
		alphabet::Symbol symbol = m_SymbolParser.parse();
		res.insert(symbol);

		next();
		token = m_GrammarLexer.token();
		if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
			break;
		}
		next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
	}
	
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::AlibException();
	}
	next();

	return res;
}

void GrammarFromStringParser::parseRules(grammar::UnknownGrammar& grammar) {
	GrammarFromStringLexer::Token token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::AlibException();
	}
	next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();

	token = m_GrammarLexer.token();
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) while(true) {
		std::vector<alphabet::Symbol> lhs;

		do {
			alphabet::Symbol symbol = m_SymbolParser.parse();
			lhs.push_back(symbol);

			next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
			token = m_GrammarLexer.token();
		} while(token.type != GrammarFromStringLexer::TokenType::MAPS_TO);

		if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) {
			throw exception::AlibException();
		}
		next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
		token = m_GrammarLexer.token();


		do {
			std::vector<alphabet::Symbol> rhs;

			if(token.type == GrammarFromStringLexer::TokenType::EPSILON) {
				next();
				token = m_GrammarLexer.token();
			} else do {
				alphabet::Symbol symbol = m_SymbolParser.parse();
				rhs.push_back(symbol);

				next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
				token = m_GrammarLexer.token();
			} while(token.type != GrammarFromStringLexer::TokenType::SET_END && token.type != GrammarFromStringLexer::TokenType::COMMA && token.type != GrammarFromStringLexer::TokenType::SEPARATOR);

			grammar.addRule(UnknownRule{ lhs, rhs } );

			if(token.type == GrammarFromStringLexer::TokenType::SET_END || token.type == GrammarFromStringLexer::TokenType::COMMA)
				break;
			else
				next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
			token = m_GrammarLexer.token();
		} while(true);

		if(token.type == GrammarFromStringLexer::TokenType::SET_END)
			break;
		else
			next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
		token = m_GrammarLexer.token();
	}
	
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::AlibException();
	}
	next();

}

} /* namespace grammar */
