/*
 * RightLG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "RightLG.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

namespace grammar {

RightLG::RightLG(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol) {

}

RightLG::RightLG(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* RightLG::clone() const {
	return new RightLG(*this);
}

GrammarBase* RightLG::plunder() && {
	return new RightLG(std::move(*this));
}

bool RightLG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >> >& rule : rules) {
		for(const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >& rhsTmp : rule.second)
			if(rhsTmp.is<std::vector<alphabet::Symbol>>()) {
				const std::vector<alphabet::Symbol>& rhs = rhsTmp.get<std::vector<alphabet::Symbol>>();

				if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
					throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
			} else {
				const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rhs = rhsTmp.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();

				if(std::find(rhs.first.begin(), rhs.first.end(), symbol) != rhs.first.end())
					throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
			}
	}

	return terminalAlphabet.erase(symbol);
}

bool RightLG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >> >& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >& rhsTmp : rule.second)
			if(rhsTmp.is<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>()) {
				const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rhs = rhsTmp.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();

				if(rhs.second == symbol)
					throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
			}
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool RightLG::addRule(const alphabet::Symbol& leftHandSide, const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >& rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	if(rightHandSide.is<std::vector<alphabet::Symbol>>()) {
		const std::vector<alphabet::Symbol>& rhs = rightHandSide.get<std::vector<alphabet::Symbol>>();
		for(const auto & symbol : rhs ) {
			if(terminalAlphabet.find(symbol) == terminalAlphabet.end())
				throw GrammarException("Symbol " + (std::string) symbol + " is not a terminal symbol");
		}

		return rules[leftHandSide].insert(rightHandSide).second;
	} else {
		const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rhs = rightHandSide.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();

		if(nonterminalAlphabet.find(rhs.second) == nonterminalAlphabet.end())
			throw GrammarException("Symbol " + (std::string) rhs.second + " is not a nonterminal symbol");

		for(const auto & symbol : rhs.first ) {
			if(terminalAlphabet.find(symbol) == terminalAlphabet.end())
				throw GrammarException("Symbol " + (std::string) symbol + " is not a terminal symbol");
		}

		return rules[leftHandSide].insert(rightHandSide).second;
	}
}

bool RightLG::addRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

bool RightLG::addRule(const alphabet::Symbol& leftHandSide, const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

const std::map<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >> > & RightLG::getRules() const {
	return rules;
}

bool RightLG::removeRule(const alphabet::Symbol& leftHandSide, const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

bool RightLG::removeRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool RightLG::removeRule(const alphabet::Symbol& leftHandSide, const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool RightLG::addRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		return addRule(leftHandSide, rightHandSide);
	} else if(nonterminalAlphabet.count(rightHandSide[rightHandSide.size() - 1])) {
		return addRule(leftHandSide, std::make_pair(std::vector<alphabet::Symbol>(rightHandSide.begin(), rightHandSide.end() - 1), rightHandSide[rightHandSide.size() - 1]));
	} else {
		return addRule(leftHandSide, rightHandSide);
	}
}

std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> RightLG::getRawRules() const {
	std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> res;
	for(const auto& rule : getRules()) {
		for(const auto& rhs : rule.second) {
			if(rhs.is<std::vector<alphabet::Symbol>>()) {
				res[rule.first].insert(rhs.get<std::vector<alphabet::Symbol>>());
			} else {
				const auto& rhsTuple = rhs.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();
				std::vector<alphabet::Symbol> tmp;
				tmp.insert(tmp.end(), rhsTuple.first.begin(), rhsTuple.first.end());
				tmp.push_back(rhsTuple.second);
				res[rule.first].insert(tmp);
			}
		}
	}
	return res;
}

bool RightLG::removeRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		return removeRule(leftHandSide, rightHandSide);
	} else if(nonterminalAlphabet.count(rightHandSide[rightHandSide.size() - 1])) {
		return removeRule(leftHandSide, std::make_pair(std::vector<alphabet::Symbol>(rightHandSide.begin(), rightHandSide.end() - 1), rightHandSide[rightHandSide.size() - 1]));
	} else {
		return removeRule(leftHandSide, rightHandSide);
	}
}

bool RightLG::operator==(const ObjectBase& other) const {
	return other == *this;
}

bool RightLG::operator<(const ObjectBase& other) const {
	return other > *this;
}

bool RightLG::operator>(const ObjectBase& other) const {
	return other < *this;
}

bool RightLG::operator==(const RightLG& other) const {
	return this->nonterminalAlphabet == other.nonterminalAlphabet && this->terminalAlphabet == other.terminalAlphabet && this->initialSymbol == other.initialSymbol && this->rules == other.rules;
}

bool RightLG::operator<(const RightLG& other) const {
	return std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules) < std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);
}

void RightLG::operator>>(std::ostream& out) const {
	out << "(RightLG"
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

RightLG::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace grammar */
