/*
 * LabelFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LabelFromStringParser.h"
#include "../exception/AlibException.h"
#include "StringLabel.h"
#include "IntegerLabel.h"
#include "CharacterLabel.h"
#include "LabelSetLabel.h"
#include "LabelPairLabel.h"
#include <algorithm>

namespace label {

LabelFromStringParser::LabelFromStringParser(std::istream& input) : m_Lexer(input) {

}

Label LabelFromStringParser::parse() {
	return parse(std::set<FEATURES>({FEATURES::STRING, FEATURES::CHAR, FEATURES::INTEGER, FEATURES::LABEL_SET, FEATURES::LABEL_PAIR}));
}

Label LabelFromStringParser::parse(const std::set<FEATURES>& features) {
	LabelFromStringLexer::Token token = m_Lexer.token();
	switch(token.type) {
	case LabelFromStringLexer::TokenType::TEOF:
		throw exception::AlibException("Unexpected end of file");
	case LabelFromStringLexer::TokenType::STRING:
		if(!features.count(FEATURES::STRING)) throw exception::AlibException();
		return Label(label::StringLabel(token.value));
	case LabelFromStringLexer::TokenType::CHAR:
		if(!features.count(FEATURES::CHAR)) throw exception::AlibException();
		return Label(label::CharacterLabel(token.value[0]));
	case LabelFromStringLexer::TokenType::INTEGER:
		if(!features.count(FEATURES::INTEGER)) throw exception::AlibException();
		return Label(label::IntegerLabel(std::stoi(token.value)));
	case LabelFromStringLexer::TokenType::SET_BEGIN: {
		if(!features.count(FEATURES::LABEL_SET)) throw exception::AlibException();
		next();
		token = m_Lexer.token();

		std::set<Label> labelSet;
		while(token.type != LabelFromStringLexer::TokenType::SET_END) {
			Label innerLabel = this->parse();
			labelSet.insert(innerLabel);

			next();
			token = m_Lexer.token();

			if(token.type == LabelFromStringLexer::TokenType::COMMA) {
				next();
				token = m_Lexer.token();
			} else
				break;
		}
		if(token.type != LabelFromStringLexer::TokenType::SET_END)
			throw exception::AlibException("Expected SET_END token.");
		return Label(label::LabelSetLabel(labelSet));
	}
	case LabelFromStringLexer::TokenType::PAIR_BEGIN: {
		if(!features.count(FEATURES::LABEL_PAIR)) throw exception::AlibException();
		next();
		token = m_Lexer.token();

		Label firstLabel = this->parse();
		next();
		token = m_Lexer.token();

		if(token.type == LabelFromStringLexer::TokenType::COMMA) {
			next();
			token = m_Lexer.token();
		} else
			throw exception::AlibException("Excepted COMMA token.");

		Label secondLabel = this->parse();
		next();
		token = m_Lexer.token();

		if(token.type != LabelFromStringLexer::TokenType::PAIR_END)
			throw exception::AlibException("Expected PAIR_END token.");
		return Label(label::LabelPairLabel(std::make_pair(firstLabel, secondLabel)));
	}
	case LabelFromStringLexer::TokenType::SET_END:
	case LabelFromStringLexer::TokenType::PAIR_END:
	case LabelFromStringLexer::TokenType::COMMA:
		throw exception::AlibException("Unexpected start of Label.");
	case LabelFromStringLexer::TokenType::ERROR:
		break;
	}
	throw exception::AlibException("Unrecognised input \"" + token.value + "\"");
}

bool LabelFromStringParser::next() {
	LabelFromStringLexer::Token token = m_Lexer.next().token();
	if(token.type == LabelFromStringLexer::TokenType::STRING || token.type == LabelFromStringLexer::TokenType::CHAR || token.type == LabelFromStringLexer::TokenType::INTEGER || token.type == LabelFromStringLexer::TokenType::SET_BEGIN || token.type == LabelFromStringLexer::TokenType::SET_END || token.type == LabelFromStringLexer::TokenType::PAIR_BEGIN || token.type == LabelFromStringLexer::TokenType::PAIR_END || token.type == LabelFromStringLexer::TokenType::COMMA) {
		return true;
	} else {
		return false;
	}
}

bool LabelFromStringParser::first() {
	LabelFromStringLexer::Token token = m_Lexer.next().token();
	if(token.type == LabelFromStringLexer::TokenType::STRING || token.type == LabelFromStringLexer::TokenType::CHAR || token.type == LabelFromStringLexer::TokenType::INTEGER || token.type == LabelFromStringLexer::TokenType::SET_BEGIN || token.type == LabelFromStringLexer::TokenType::PAIR_BEGIN) {
		return true;
	} else {
		return false;
	}
}

bool LabelFromStringParser::last() {
	LabelFromStringLexer::Token token = m_Lexer.next().token();
	if(token.type == LabelFromStringLexer::TokenType::TEOF) {
		return true;
	} else {
		return false;
	}
}

} /* namespace label */
