/*
 * LabelFromXMLParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LabelFromXMLParser.h"
#include "../sax/ParserException.h"
#include <algorithm>

namespace label {

Label LabelFromXMLParser::parseLabel(std::list<sax::Token>& input) const {
	return parseLabel(input, std::set<FEATURES>({FEATURES::STRING, FEATURES::CHAR, FEATURES::INTEGER, FEATURES::LABEL_SET, FEATURES::LABEL_PAIR}));
}

Label LabelFromXMLParser::parseLabel(std::list<sax::Token>& input, const std::set<FEATURES>& features) const {
	if(isToken(input, sax::Token::TokenType::START_ELEMENT, "IntegerLabel")) {
		if(!features.count(FEATURES::INTEGER)) throw exception::AlibException();
		return Label(parseIntegerLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "StringLabel")) {
		if(!features.count(FEATURES::STRING)) throw exception::AlibException();
		return Label(parseStringLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "CharacterLabel")) {
		if(!features.count(FEATURES::CHAR)) throw exception::AlibException();
		return Label(parseCharacterLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelSetLabel")) {
		if(!features.count(FEATURES::LABEL_SET)) throw exception::AlibException();
		return Label(parseLabelSetLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelPairLabel")) {
		if(!features.count(FEATURES::LABEL_PAIR)) throw exception::AlibException();
		return Label(parseLabelPairLabel(input));
	} else {
		throw sax::ParserException(sax::Token("IntegerLabel, StringLabel, CharacterLabel, LabelSetLabel, LabelPairLabel", sax::Token::TokenType::START_ELEMENT), input.front());
	}
}

bool LabelFromXMLParser::first(std::list<sax::Token>& input) const {
	if(isToken(input, sax::Token::TokenType::START_ELEMENT, "IntegerLabel") || isToken(input, sax::Token::TokenType::START_ELEMENT, "StringLabel") || isToken(input, sax::Token::TokenType::START_ELEMENT, "CharacterLabel") || isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelSetLabel") || isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelPairLabel")) {
		return true;
	} else {
		return false;
	}
}

IntegerLabel LabelFromXMLParser::parseIntegerLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "IntegerLabel");
	IntegerLabel data(std::stoi(popTokenData(input, sax::Token::TokenType::CHARACTER)));
	popToken(input, sax::Token::TokenType::END_ELEMENT, "IntegerLabel");
	return data;
}

CharacterLabel LabelFromXMLParser::parseCharacterLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "CharacterLabel");
	CharacterLabel data(popTokenData(input, sax::Token::TokenType::CHARACTER)[0]);
	popToken(input, sax::Token::TokenType::END_ELEMENT, "CharacterLabel");
	return data;
}

StringLabel LabelFromXMLParser::parseStringLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "StringLabel");
	StringLabel data(popTokenData(input, sax::Token::TokenType::CHARACTER));
	popToken(input, sax::Token::TokenType::END_ELEMENT, "StringLabel");
	return data;
}

LabelSetLabel LabelFromXMLParser::parseLabelSetLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "LabelSetLabel");
	std::set<Label> labels;
	while(isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		labels.insert(parseLabel(input));
	}
	LabelSetLabel data(labels);
	popToken(input, sax::Token::TokenType::END_ELEMENT, "LabelSetLabel");
	return data;
}

LabelPairLabel LabelFromXMLParser::parseLabelPairLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "LabelPairLabel");
	
	Label firstLabel = parseLabel(input);
	Label secondLabel = parseLabel(input);

	LabelPairLabel data(std::make_pair(firstLabel, secondLabel));
	popToken(input, sax::Token::TokenType::END_ELEMENT, "LabelPairLabel");
	return data;
}

} /* namespace label */
