/*
 * FormalRegExp.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "FormalRegExp.h"
#include "../../exception/AlibException.h"
#include "FormalRegExpEmpty.h"
#include "FormalRegExpSymbol.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../../std/set.hpp"
#include "../unbounded/UnboundedRegExp.h"

namespace regexp {

FormalRegExp::FormalRegExp() {
	this->regExp = new FormalRegExpEmpty();
}

FormalRegExp::FormalRegExp(const UnboundedRegExp& other) : alphabet(other.getAlphabet()) {
	this->regExp = NULL;
	FormalRegExpElement* element = other.getRegExp().cloneAsFormal();
	setRegExp(std::move(*element));
	delete element;
}

FormalRegExp::FormalRegExp(const std::set<alphabet::Symbol>& alphabet, const FormalRegExpElement& regExp) : alphabet(alphabet) {
	this->regExp = NULL;
	setRegExp(regExp);
}

FormalRegExp::FormalRegExp(std::set<alphabet::Symbol>&& alphabet, FormalRegExpElement&& regExp) : alphabet(std::move(alphabet)) {
	this->regExp = NULL;
	setRegExp(std::move(regExp));
}

FormalRegExp::FormalRegExp(const FormalRegExpElement& regExp) {
	regExp.computeMinimalAlphabet(alphabet);
	this->regExp = NULL;
	setRegExp(regExp);
}

FormalRegExp::FormalRegExp(FormalRegExpElement&& regExp) {
	regExp.computeMinimalAlphabet(alphabet);
	this->regExp = NULL;
	setRegExp(std::move(regExp));
}

FormalRegExp::FormalRegExp(const FormalRegExp& other) : regExp(other.regExp->clone()), alphabet(other.alphabet) {
	this->regExp->attachRegExp(this);
}

FormalRegExp::FormalRegExp(FormalRegExp&& other) noexcept : regExp(other.regExp), alphabet(std::move(other.alphabet) ) {
	this->regExp->attachRegExp(this);
	other.regExp = NULL;
}

RegExpBase* FormalRegExp::clone() const {
	return new FormalRegExp(*this);
}

RegExpBase* FormalRegExp::plunder() && {
	return new FormalRegExp(std::move(*this));
}

FormalRegExp& FormalRegExp::operator=(const FormalRegExp& other) {
	if (this == &other) {
		return *this;
	}

	*this = FormalRegExp(other);

	return *this;
}

FormalRegExp& FormalRegExp::operator=(FormalRegExp&& other) noexcept {
	std::swap(this->regExp, other.regExp);
	std::swap(this->alphabet, other.alphabet);
	return *this;
}

FormalRegExp::~FormalRegExp() noexcept {
	delete regExp;
}

const FormalRegExpElement& FormalRegExp::getRegExp() const {
	return *regExp;
}

FormalRegExpElement& FormalRegExp::getRegExp() {
	return *regExp;
}

void FormalRegExp::setRegExp(const FormalRegExpElement& regExp) {
	delete this->regExp;
	this->regExp = regExp.clone();
	if(!this->regExp->attachRegExp(this))
		throw exception::AlibException("Input symbols not in the alphabet.");
}

void FormalRegExp::setRegExp(FormalRegExpElement&& regExp) {
	delete this->regExp;
	this->regExp = std::move(regExp).plunder();
	if(!this->regExp->attachRegExp(this))
		throw exception::AlibException("Input symbols not in the alphabet.");
}

const std::set<alphabet::Symbol>& FormalRegExp::getAlphabet() const {
	return alphabet;
}

bool FormalRegExp::addSymbolToAlphabet(const alphabet::Symbol & symbol) {
	return alphabet.insert(symbol).second;
}

void FormalRegExp::setAlphabet(const std::set<alphabet::Symbol> & symbols) {
	std::set<alphabet::Symbol> minimalAlphabet;
	this->regExp->computeMinimalAlphabet(minimalAlphabet);
	std::set<alphabet::Symbol> removedSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), symbols.begin(), symbols.end(), std::inserter(removedSymbols, removedSymbols.end()));

	if(removedSymbols.size() > 0)
		throw exception::AlibException("Input symbols are used.");

	this->alphabet = symbols;
}

bool FormalRegExp::removeSymbolFromAlphabet(const alphabet::Symbol & symbol) {
	if(this->regExp->testSymbol(symbol))
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	return alphabet.erase(symbol);
}

bool FormalRegExp::isEmpty() const {
	return regExp->isEmpty();
}

bool FormalRegExp::containsEmptyString() const {
	return regExp->containsEmptyString();
}

void FormalRegExp::operator >>(std::ostream& out) const {
	out << "(FormalRegExp " << *(this->regExp) << ")";
}

bool FormalRegExp::operator<(const FormalRegExp& other) const {
	return std::tie(*regExp, alphabet) < std::tie(*other.regExp, other.alphabet);
}

bool FormalRegExp::operator==(const FormalRegExp& other) const {
	return *regExp == *other.regExp && this->alphabet == other.alphabet;
}

bool FormalRegExp::operator<(const ObjectBase& other) const {
	return other > *this;
}

bool FormalRegExp::operator ==(const ObjectBase& other) const {
	return other == *this;
}

bool FormalRegExp::operator >(const ObjectBase& other) const {
	return other < *this;
}

FormalRegExp::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace regexp */
