/*
 * FormalRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_ELEMENT_H_
#define FORMAL_REG_EXP_ELEMENT_H_

#include "../../std/visitor.hpp"
#include "../../alphabet/Symbol.h"
#include "FormalRegExp.h"
#include <set>

namespace regexp {

class FormalRegExp;

class FormalRegExpAlternation;
class FormalRegExpConcatenation;
class FormalRegExpIteration;
class FormalRegExpSymbol;
class FormalRegExpEmpty;
class FormalRegExpEpsilon;

class UnboundedRegExpElement;

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
class FormalRegExpElement : public std::acceptor_base<FormalRegExpAlternation, FormalRegExpConcatenation, FormalRegExpIteration, FormalRegExpSymbol, FormalRegExpEmpty, FormalRegExpEpsilon> {
protected:
	/* 
	 * Parent regexp contanining this instance of RegExpElement
	 */
	const FormalRegExp * parentRegExp;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Attaches the regexp to this instance and all its childs
	 * @param regexp parent regexp to attach as parent
	 * @return true if symbols used in regexp element are in the regexp's alphabet
	 */
	virtual bool attachRegExp ( const FormalRegExp * regexp ) = 0;
	
	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const = 0;
	
public:
	FormalRegExpElement();
	
	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual FormalRegExpElement* clone() const = 0;
	
	virtual FormalRegExpElement* plunder() && = 0;

	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual UnboundedRegExpElement* cloneAsUnbounded() const = 0;

	virtual ~FormalRegExpElement() noexcept;

	// RegExpEmpty < RegExpEpsilon < RegExpSymbol < RegExpIteration < RegExpAlternation < RegExpConcatenation
	virtual bool operator<(const FormalRegExpElement&) const = 0;
	virtual bool operator==(const FormalRegExpElement&) const = 0;
	virtual bool operator>(const FormalRegExpElement&) const = 0;

	virtual bool operator>=(const FormalRegExpElement&) const;
	virtual bool operator<=(const FormalRegExpElement&) const;
	virtual bool operator!=(const FormalRegExpElement&) const;

	virtual bool operator<(const FormalRegExpConcatenation&) const;
	virtual bool operator<(const FormalRegExpAlternation&) const;
	virtual bool operator<(const FormalRegExpIteration&) const;
	virtual bool operator<(const FormalRegExpSymbol&) const;
	virtual bool operator<(const FormalRegExpEpsilon&) const;
	virtual bool operator<(const FormalRegExpEmpty&) const;

	virtual bool operator==(const FormalRegExpConcatenation&) const;
	virtual bool operator==(const FormalRegExpAlternation&) const;
	virtual bool operator==(const FormalRegExpIteration&) const;
	virtual bool operator==(const FormalRegExpSymbol&) const;
	virtual bool operator==(const FormalRegExpEpsilon&) const;
	virtual bool operator==(const FormalRegExpEmpty&) const;
	
	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 */
	virtual void operator>>(std::ostream& out) const = 0;
	
	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	friend std::ostream& operator<<(std::ostream& out, const FormalRegExpElement& regexp);

	/**
	 * @return true if this subtree of regexp matches empty string (epsilon)
	 */
	virtual bool containsEmptyString() const = 0;

	/**
	 * @return true if this subtree describes empty language
	 */
	virtual bool isEmpty() const = 0;
	
	friend class FormalRegExp;
	
	friend class FormalRegExpAlternation;
	friend class FormalRegExpConcatenation;
	friend class FormalRegExpIteration;
	friend class FromalRegExpSymbol;
	friend class FormalRegExpEmpty;
	friend class FormalRegExpEpsilon;
};

} /* namespace regexp */

#endif /* FORMAL_REG_EXP_ELEMENT_H_ */
