/*
 * UnboundedRegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_H_
#define UNBOUNDED_REG_EXP_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "UnboundedRegExpElement.h"
#include "../../std/visitor.hpp"
#include "../RegExpBase.h"

namespace regexp {

class UnboundedRegExpElement;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class UnboundedRegExp : public std::acceptor<UnboundedRegExp, VisitableRegExpBase, std::acceptor<UnboundedRegExp, alib::VisitableObjectBase, RegExpBase> > {
protected:
	UnboundedRegExpElement* regExp;
	
	std::set<alphabet::Symbol> alphabet;

public:
	/**
	 * @copydoc UnboundedRegExpElement::clone() const
	 */
	virtual RegExpBase* clone() const;
	
	/**
	 * @copydoc UnboundedRegExpElement::plunder() const
	 */
	virtual RegExpBase* plunder() &&;

	UnboundedRegExp();
	explicit UnboundedRegExp(const FormalRegExp& other);
	UnboundedRegExp(const std::set<alphabet::Symbol>& alphabet, const UnboundedRegExpElement& regExp);
	UnboundedRegExp(std::set<alphabet::Symbol>&& alphabet, UnboundedRegExpElement&& regExp);
	explicit UnboundedRegExp(const UnboundedRegExpElement& regExp);
	explicit UnboundedRegExp(UnboundedRegExpElement&& regExp);

	/**
	 * Copy constructor.
	 * @param other RegExp to copy
	 */
	UnboundedRegExp(const UnboundedRegExp& other);
	UnboundedRegExp(UnboundedRegExp&& other) noexcept;
	UnboundedRegExp& operator =(const UnboundedRegExp& other);
	UnboundedRegExp& operator =(UnboundedRegExp&& other) noexcept;
	~UnboundedRegExp() noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const UnboundedRegExpElement& getRegExp() const;
	
	/**
	 * @return Root node of the regular expression tree
	 */
	UnboundedRegExpElement& getRegExp();

	/**
	 * Sets the root node of the regular expression tree. Doesn't perform copy of the regExp param,
	 * just stores it!
	 * @param regExp root node to set
	 */
	void setRegExp(const UnboundedRegExpElement& regExp);
	
	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp(UnboundedRegExpElement&& regExp);
	
	/**
	 * Gets alphabet symbols used in RegExp.
	 * @return set of alphabet symbols used in regexp.
	 */
	const std::set<alphabet::Symbol>& getAlphabet() const;
	
	/**
	 * Adds symbol to the alphabet available in the regular expression
	 * @param symbol new symbol added to the alphabet
	 */
	bool addSymbolToAlphabet(const alphabet::Symbol & symbol);

	/**
	 * Sets the alphabet of the regular expression
	 * @param symbols new alphabet
	 */
	void setAlphabet(const std::set<alphabet::Symbol>& symbols);
	
	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet(const alphabet::Symbol & symbol);
	
	/**
	 * @return true if regexp represents empty language
	 */
	bool isEmpty() const;

	/**
	 * @return true if regexp matches empty string (epsilon)
	 */
	bool containsEmptyString() const;

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	virtual void operator>>(std::ostream& out) const;

	virtual bool operator <(const alib::ObjectBase& other) const;
	virtual bool operator ==(const alib::ObjectBase& other) const;
	virtual bool operator >(const alib::ObjectBase& other) const;

	virtual bool operator==(const UnboundedRegExp& other) const;

	virtual bool operator<(const UnboundedRegExp& other) const;

	virtual operator std::string() const;

	virtual int selfTypeId() const {
		return typeId(*this);
	}
};

} /* namespace regexp */

#endif /* UNBOUNDED_REG_EXP_H_ */
