/*
 * UnboundedRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_ELEMENT_H_
#define UNBOUNDED_REG_EXP_ELEMENT_H_

#include "../../std/visitor.hpp"
#include "../../alphabet/Symbol.h"
#include "UnboundedRegExp.h"
#include <set>

namespace regexp {

class UnboundedRegExp;

class UnboundedRegExpAlternation;
class UnboundedRegExpConcatenation;
class UnboundedRegExpIteration;
class UnboundedRegExpSymbol;
class UnboundedRegExpEmpty;
class UnboundedRegExpEpsilon;

class FormalRegExpElement;

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
class UnboundedRegExpElement : public std::acceptor_base<UnboundedRegExpAlternation, UnboundedRegExpConcatenation, UnboundedRegExpIteration, UnboundedRegExpSymbol, UnboundedRegExpEmpty, UnboundedRegExpEpsilon> {
protected:
	/**
	 * Parent regexp contanining this instance of RegExpElement
	 */
	const UnboundedRegExp * parentRegExp;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Attaches the regexp to this instance and all its childs
	 * @param regexp parent regexp to attach as parent
	 * @return true if symbols used in regexp element are in the regexp's alphabet
	 */
	virtual bool attachRegExp ( const UnboundedRegExp * regexp ) = 0;
	
	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const = 0;
	
public:
	UnboundedRegExpElement();
	
	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual UnboundedRegExpElement* clone() const = 0;
	
	virtual UnboundedRegExpElement* plunder() && = 0;

	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual FormalRegExpElement* cloneAsFormal() const = 0;

	virtual ~UnboundedRegExpElement() noexcept;

	// RegExpEmpty < RegExpEpsilon < RegExpSymbol < RegExpIteration < RegExpAlternation < RegExpConcatenation
	virtual bool operator<(const UnboundedRegExpElement&) const = 0;
	virtual bool operator==(const UnboundedRegExpElement&) const = 0;
	virtual bool operator>(const UnboundedRegExpElement&) const = 0;

	virtual bool operator>=(const UnboundedRegExpElement&) const;
	virtual bool operator<=(const UnboundedRegExpElement&) const;
	virtual bool operator!=(const UnboundedRegExpElement&) const;

	virtual bool operator<(const UnboundedRegExpConcatenation&) const;
	virtual bool operator<(const UnboundedRegExpAlternation&) const;
	virtual bool operator<(const UnboundedRegExpIteration&) const;
	virtual bool operator<(const UnboundedRegExpSymbol&) const;
	virtual bool operator<(const UnboundedRegExpEpsilon&) const;
	virtual bool operator<(const UnboundedRegExpEmpty&) const;

	virtual bool operator==(const UnboundedRegExpConcatenation&) const;
	virtual bool operator==(const UnboundedRegExpAlternation&) const;
	virtual bool operator==(const UnboundedRegExpIteration&) const;
	virtual bool operator==(const UnboundedRegExpSymbol&) const;
	virtual bool operator==(const UnboundedRegExpEpsilon&) const;
	virtual bool operator==(const UnboundedRegExpEmpty&) const;
	
	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 */
	virtual void operator>>(std::ostream& out) const = 0;
	
	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	friend std::ostream& operator<<(std::ostream& out, const UnboundedRegExpElement& regexp);

	/**
	 * @return true if this subtree of regexp matches empty string (epsilon)
	 */
	virtual bool containsEmptyString() const = 0;

	/**
	 * @return true if this subtree describes empty language
	 */
	virtual bool isEmpty() const = 0;
	
	friend class UnboundedRegExp;
	
	friend class UnboundedRegExpAlternation;
	friend class UnboundedRegExpConcatenation;
	friend class UnboundedRegExpIteration;
	friend class FromalRegExpSymbol;
	friend class UnboundedRegExpEmpty;
	friend class UnboundedRegExpEpsilon;
};

} /* namespace regexp */

#endif /* UNBOUNDED_REG_EXP_ELEMENT_H_ */
