#include <iostream>
#include <getopt.h>

#include <factory/DataFactory.hpp>
#include <exception/AlibException.h>

#include "trim/grammar/TrimCFG.h"
#include "trim/automaton/TrimFSM.h"
#include "regexp/RegExpOptimize.h"

void help( void ) {
	std::cout << "atrim 0.01" << std::endl;
	std::cout << "Removes unreachable and useless states from FSM, productive and unreachable nonterminals from CFG. Simplifies representation of RE" << std::endl;
	std::cout << "Usage: atrim [-u] [-r]" << std::endl << std::endl;
	std::cout << "If neither --useless nor --unreachable option is used, both useless and unreachable states (or symbols) are removed." << std::endl;
	std::cout << std::endl;
	std::cout << " -u, --useless \t Removes useless states. (works with FSM only)" << std::endl;
	std::cout << " -r, --unreachable \t Removes unreachable states. (works with FSM or CFG)" << std::endl;
	std::cout << " -s, --simplify \t Simplifies representation. (works with RE)" << std::endl;
	std::cout << " -h, --help \t shows this." << std::endl;
	std::cout << std::endl;
}

template<typename T>
grammar::GrammarBase* dynamicTrim(const grammar::Grammar& g, bool del_unreachable, bool del_useless) {
	const T* gp = dynamic_cast<const T*>(&g.getData());

	if(gp) {
		T res(*gp);
		if( del_unreachable )
			res = trim::TrimCFG::removeUnreachableSymbols( res );
		if( del_useless )
			res = trim::TrimCFG::removeUnproductiveSymbols( res );
		return std::move(res).plunder();
	} else {
		return NULL;
	}
}

grammar::Grammar trimGrammar(const grammar::Grammar& g, bool del_unreachable, bool del_useless) {
	grammar::GrammarBase* res = NULL;

	res = dynamicTrim<grammar::CFG>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	res = dynamicTrim<grammar::EpsilonFreeCFG>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	res = dynamicTrim<grammar::GNF>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	res = dynamicTrim<grammar::CNF>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	res = dynamicTrim<grammar::LeftLG>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	res = dynamicTrim<grammar::LeftRG>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	res = dynamicTrim<grammar::RightLG>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	res = dynamicTrim<grammar::RightRG>(g, del_unreachable, del_useless);
	if(res) return grammar::Grammar(*res);

	throw exception::AlibException("Unsupported grammar type");
}

template<typename T>
automaton::AutomatonBase* dynamicTrim(const automaton::Automaton& g, bool del_unreachable, bool del_useless) {
	const T* gp = dynamic_cast<const T*>(&g.getData());

	if(gp) {
		T res(*gp);
		if( del_unreachable )
			res = trim::TrimFSM::removeUnreachableStates( res );
		if( del_useless )
			res = trim::TrimFSM::removeUselessStates( res );
		return std::move(res).plunder();
	} else {
		return NULL;
	}
}

automaton::Automaton trimAutomaton(const automaton::Automaton& g, bool del_unreachable, bool del_useless) {
	automaton::AutomatonBase* res = NULL;

	res = dynamicTrim<automaton::DFA>(g, del_unreachable, del_useless);
	if(res) return automaton::Automaton(*res);

	res = dynamicTrim<automaton::NFA>(g, del_unreachable, del_useless);
	if(res) return automaton::Automaton(*res);

	res = dynamicTrim<automaton::EpsilonNFA>(g, del_unreachable, del_useless);
	if(res) return automaton::Automaton(*res);

	res = dynamicTrim<automaton::CompactNFA>(g, del_unreachable, del_useless);
	if(res) return automaton::Automaton(*res);

	res = dynamicTrim<automaton::ExtendedNFA>(g, del_unreachable, del_useless);
	if(res) return automaton::Automaton(*res);

	throw exception::AlibException("Unsupported automaton type");
}

template<typename T>
regexp::RegExpBase* dynamicOptimize(const regexp::RegExp& r) {
	const T* rp = dynamic_cast<const T*>(&r.getData());

	if(rp) {
		T res(*rp);
		regexp::RegExpOptimize opt;
		res = opt.optimize( res );
		return std::move(res).plunder();
	} else {
		return NULL;
	}
}

regexp::RegExp optimizeRegExp(const regexp::RegExp& r) {
	regexp::RegExpBase* res = NULL;

	res = dynamicOptimize<regexp::UnboundedRegExp>(r);
	if(res) return regexp::RegExp(*res);

	res = dynamicOptimize<regexp::FormalRegExp>(r);
	if(res) return regexp::RegExp(*res);

	throw exception::AlibException("Unsupported automaton type");
}

int main(int argc, char* argv[]) {
	bool del_useless = false, del_unreachables = false;

	static struct option long_options[] = {
		{"help",		no_argument,	NULL,	'h'},
		{"useless",		no_argument,	NULL,	'u'},
		{"unreachable", 	no_argument,	NULL,	'r'},
		{0,			0,		0,	0}
	};

	int long_index = 0, opt = 0;

	while( ( opt = getopt_long( argc, argv, "hru", long_options, & long_index ) ) != -1 ) {
		switch( opt ) {
		case 'u':
			del_useless = true;
			break;
		case 'r':
			del_unreachables = true;
			break;

		case 'v':
		case 'h':
		default:
			help( );
			return 0;
		}
	}

	std::list<sax::Token> tokens;
	if(optind == argc) {
		sax::SaxParseInterface::parseStdin(tokens);
	} else {
		sax::SaxParseInterface::parseFile(argv[optind], tokens);
	}

	if( ! del_useless && ! del_unreachables )
		del_useless = del_unreachables = true;

	try {
		if( alib::FromXMLParsers::automatonParser.first(tokens)) {
			automaton::Automaton automaton = alib::DataFactory::fromTokens<automaton::Automaton>(tokens);

			automaton::Automaton res = trimAutomaton(automaton, del_unreachables, del_useless );

			alib::DataFactory::toStdout( res );
			return 0;
		} else if( alib::FromXMLParsers::grammarParser.first(tokens)) {
			grammar::Grammar grammar = alib::DataFactory::fromTokens<grammar::Grammar>(tokens);

			grammar::Grammar res = trimGrammar(grammar, del_unreachables, del_useless );

			alib::DataFactory::toStdout( res );
			return 0;
		} else {
			throw exception::AlibException( "Input not recognized." );
			return 1;
		}
	} catch( const exception::AlibException & e ) {
		alib::DataFactory::toStdout( e );
		return 1;
	} catch (...) {
		return 127;
	}
}
